from datetime import datetime
from django.db.models import Max
from django.db.models.query_utils import Q
from django.shortcuts import render_to_response
from django.views.decorators.cache import cache_page
from mywlcg.base.views import path_prefix
from mywlcg.mdb.models import MetricData, MetricDataLatest


@cache_page(600)
def index(request):
    return render_to_response('metric_index.html')


@cache_page(600)
def json(request):
    # Datatables GET parameters
    # http://www.datatables.net/usage/server-side
    iDisplayStart = int(request.GET.get("iDisplayStart") or 0)
    iDisplayLength = int(request.GET.get("iDisplayLength") or 10)
    #iColumns = int(request.GET.get("iColumns") or -1)
    sSearch = request.GET.get("sSearch")
    #bEscapeRegex = request.GET.get("bEscapeRegex")
    #iSortingCols = int(request.GET.get("iSortingCols") or -1)
    sEcho = int(request.GET.get("sEcho") or -1)

    # Initial setup
    queryset = MetricDataLatest.objects
    iTotalRecords = queryset.count()

    # Search
    if sSearch:
        queryset = queryset.filter(service__hostname__icontains=sSearch)

    # Extracting results
    iTotalDisplayRecords = queryset.count()
    metrics = queryset.values_list(
        'id', 'service__hostname',
        'metric__name', 'check_time', 'metricstatus__description'
    )[iDisplayStart:iDisplayStart + iDisplayLength]
    results = {
        'metrics': metrics,
        'iTotalRecords': iTotalRecords,
        'iTotalDisplayRecords': iTotalDisplayRecords,
        'sEcho': sEcho,
    }
    path_prefix(request, results)
    return render_to_response('metric_json', results)


def _get_metricdata_object(metricdata_id, model):
    return model.objects.select_related(
        'service__hostname', 'metric__name', 'metricstatus__description'
    ).get(id=metricdata_id)


def _details(request, mdata, prefix='../../'):
    metric_details = 'no details'
    if mdata.metricdetail_id:
        metric_details = mdata.metricdetail.detail.decode('string_escape')
    elif mdata.summarydata:
        metric_details = mdata.summarydata.decode('string_escape')

    permalink = "%s%s/%s/%s/%s/" % (
        prefix,
        mdata.vo_id or '',
        mdata.service_id,
        mdata.metric_id,
        mdata.check_time,
    )
    results = [
        mdata.service.hostname,
        mdata.metric.name,
        datetime.utcfromtimestamp(int(mdata.check_time)).strftime('%x %X UTC'),
        mdata.metricstatus.description,
        mdata.vo.voname or '',
        metric_details,
        permalink
    ]
    return results


@cache_page(600)
def metricdata(request, metricdata_id):
    return render_to_response('metric_details.html', {
        'results': _details(request, _get_metricdata_object(metricdata_id, MetricData)),
    })


@cache_page(600)
def metricdatalatest(request, metricdata_id):
    return render_to_response('metric_details.html', {
        'results': _details(request, _get_metricdata_object(metricdata_id, MetricDataLatest)),
    })


@cache_page(600)
def metricdatalatestpopup(request, metricdata_id):
    return render_to_response('metric_details_popup.html', {
        'results': _details(request, _get_metricdata_object(metricdata_id, MetricDataLatest, '../../metric/')),
    })


@cache_page(600)
def metricdata2(request, service_id, metric_id, timestamp):
    mdata = MetricData.objects.select_related(
        'service__hostname', 'metric__name', 'metricstatus__description'
    ).get(
        service=service_id, metric=metric_id, check_time=timestamp
    )
    return render_to_response('metric_details.html', {
        'results': _details(request, mdata, '../../../'),
    })


@cache_page(600)
def metricdata2popup(request, vo_id, service_id, metric_id, timestamp):
    timestamp = int(timestamp)
    qs = MetricData.objects.filter(
        Q(vo__isnull=True) | Q(vo=vo_id),
        service=service_id, metric=metric_id
    )
    # attempt to find the metric details in the last hour
    data = qs.filter(check_time__range=(timestamp - 3600, timestamp))
    if not data:
        # search the whole history
        data = qs.filter(check_time__lte=timestamp)
    max_check_time = data.aggregate(Max('check_time'))['check_time__max']
    mdata = data.filter(check_time=max_check_time)[0]
    return render_to_response('metric_details_popup.html', {
        'results': _details(request, mdata, request.META.get('SCRIPT_NAME') + '/metric/popup/'),
    })
