import time
import simplejson as json
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta

from django.db.models import Max
from django.shortcuts import render_to_response
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse, Http404
from django.utils.datastructures import SortedDict

from mywlcg.base.views import path_prefix
from mywlcg.mdb.models import SnapshotTopologyView, Profile, ACE_D_SITE_AVAILABILITY


MAX_N_HISTORY_MONTHS = 6
HISTORY_RANGE = (0, MAX_N_HISTORY_MONTHS)
POEM_PROFILE_URL = "https://grid-monitoring.cern.ch/poem/admin/poem/profile/?name=%s"


def all_vos():
    return SnapshotTopologyView.objects.filter(
        tier__in=('Tier-0', 'Tier-1')
    ).values_list('voname', flat=True).distinct().order_by('voname')


def all_sites():
    all_sites = SnapshotTopologyView.objects.filter(
        tier__in=('Tier-0', 'Tier-1')
    ).values('physitename').distinct()
    return sorted([s['physitename'] for s in all_sites], key=lambda s: s.upper())


def get_profile(name, vo):
    try:
        p = Profile.objects.get(name=name, atp_vo__voname=vo)
        # annotate the profile with its URL to poem
        p.profile_admin_url = POEM_PROFILE_URL % p.name
        return p
    except ObjectDoesNotExist:
        return None

VO_CRITICAL_PROFILES = {
    'ops': get_profile('WLCG_CREAM_LCGCE_CRITICAL', 'ops'),
    'atlas': get_profile('ATLAS_CRITICAL', 'atlas'),
    'alice': get_profile('ALICE_CRITICAL', 'alice'),
    'cms': get_profile('CMS_CRITICAL', 'cms'),
    'lhcb': get_profile('LHCb_CRITICAL', 'lhcb')
}


def get_dates(month, history):
    """
    * latest with no history returns the last 30 days.
    * latest with history returns the complete past months and up to the
      day of the current month.
    * full months return the complete period.
    * if history, the whole period is included in the last place
    """
    if month == 'latest':
        end = datetime.utcnow().date()
        if history == 0:
            return end - timedelta(days=30), end
        start = end - relativedelta(months=history)
        return start.replace(day=1), end
    end = (month + relativedelta(months=1)).replace(day=1)
    return end - relativedelta(months=history + 1), end


def get_tier_sites(voname=None, sitename=None, sortby='logsitename'):
    """
    returns (physical sitename, voname, group name, group_id) for all tier 0/1s in
    voname or the specific sites with provided sitename
    """
    extra = {}
    if voname is not None:
        if type(voname) not in (str, unicode):
            extra['voname__in'] = voname
        else:
            extra['voname'] = voname
    if sitename is not None:
        if type(sitename) not in (str, unicode):
            extra['physitename__in'] = sitename
        else:
            extra['physitename'] = sitename

    snapts = SnapshotTopologyView.objects.aggregate(Max('snapshotts'))['snapshotts__max']
    fields = ('physitename', 'voname', 'logsitename', 'loggroupingid')
    sites = SnapshotTopologyView.objects.filter(
        tier__in=('Tier-0', 'Tier-1'),
        snapshotts=snapts,
        **extra
    ).values(*fields).distinct()
    return sorted(sites, key=lambda o: o[sortby].lower())


def date_periods(start, end, history=0):
    """
    generator for the different (start, end, title) times in the report.
    """
    fmt = lambda d: d.strftime('%Y-%m-%dT%H:%M:%S')
    rend = end

    # if ending today, latest period is selected
    if end == datetime.utcnow().date():
        rstart = end - relativedelta(months=1)
        yield (fmt(rstart), fmt(end), 'Last 30 days')
        if history == 0:
            raise StopIteration
        rend = end.replace(day=1)

    # iterate over past months
    rstart = end - relativedelta(months=1)
    while rstart >= start:
        yield (fmt(rstart), fmt(rend), rstart.strftime("%B %Y"))
        rend = rstart
        rstart = rend - relativedelta(months=1)

    if history:  # provide a full picture range
        yield(
            fmt(start), fmt(end),
            '%s - %s' % (start.strftime("%d %B %Y"), end.strftime("%d %B %Y"))
        )


def plot_groups(group_by, vos, sites):
    """
    returns an iterable with each of the plots grouped by vo or site, which
    contains an iterable for each site plot to be displayed
    """
    def vo_plots(vo, vos, sites):
        """generate a site tuple for each vo and its profile"""
        p = VO_CRITICAL_PROFILES.get(vo)
        if not p:
            raise StopIteration
        for site in get_tier_sites(voname=vo, sitename=sites):
            title = "%s (%s)" % (site['logsitename'], vo)
            yield (site, vo, p.id, title)

    def site_plots(site, vos, sites):
        """generate a vo/profile tuple for each site"""
        for site in get_tier_sites(sitename=site, voname=vos):
            vo = site.get('voname')
            title = "%s (%s)" % (site['logsitename'], vo)
            profile = VO_CRITICAL_PROFILES.get(vo)
            if profile:
                yield(site, vo, profile.id, title)

    # set who to iter over the whole set, but disable its constraint
    if vos == 'all':
        vos = all_vos()
    if sites == 'all':
        sites = all_sites()

    who = vos
    plots = vo_plots
    if group_by == 'site':
        who = sites
        plots = site_plots

    for w in who:
        yield {
            'title': w.upper(),
            'plots': list(plots(w, vos, sites)),
            'who': w,
            'profile': VO_CRITICAL_PROFILES.get(str(w))
        }


def index(request, printable=False):
    """
    Prepares the plots and renders the dashboard
    """
    if 'myegi' in request.path:
        raise Http404

    values = request.GET.get('values')
    if values not in ('availability', 'reliability'):
        values = 'availability'

    try:
        month = request.GET.get('month', 'latest')
        if month != 'latest':
            month = datetime(*(time.strptime(month, '%Y-%m')[0:6]))
    except ValueError:
        month = datetime.utcnow().strftime('%Y-%m')

    try:
        history = int(request.GET.get('history', 0))
        vmin, vmax = HISTORY_RANGE
        if history < vmin and history > vmax:
            history = 0
    except ValueError:
        history = 0

    start, end = get_dates(month, history)

    # the template needs always a string to work both with months and 'latest'
    if type(month) is datetime:
        month = month.strftime('%Y-%m')

    groupby = request.GET.get('groupby', 'vo')
    if groupby not in ('vo', 'site'):
        groupby = 'vo'

    def value_or_split(orig, value='all', delim=','):
        if orig != value:
            return orig.split(delim)
        return value
    vos = value_or_split(request.GET.get('vos', 'all'))
    sites = value_or_split(request.GET.get('sites', 'all'))

    context = {
        # UI data
        'tab': 'tiers',
        'values': values,
        'groupby': groupby,
        'vos': request.GET.get('vos', 'all'),
        'sites': request.GET.get('sites', 'all'),
        'month': month,
        'history': history,
        'nhistory': xrange(MAX_N_HISTORY_MONTHS),

        # plot groups data
        'periods': date_periods(start, end, history),
        'groups': list(plot_groups(groupby, vos, sites)),
    }
    path_prefix(request, context)
    if printable:
        return render_to_response('reports_tiers_print.html', context)
    return render_to_response('reports_tiers.html', context)


def site_vos(request):
    """
    returns a json dictionary containing all sites for the provided VOs
    """
    vos = request.GET.get('vos')
    if vos:
        vos = map(lambda v: v.lower(), vos.split(','))
    else:
        vos = all_vos()

    sites = SnapshotTopologyView.objects.filter(
        tier__in=('Tier-0', 'Tier-1'),
        voname__in=vos
    ).values('physitename', 'voname').distinct()

    site_vos = {}
    for s in sites:
        vo, name = s['voname'], s['physitename']
        if name in site_vos:
            site_vos[name].add(vo)
        else:
            site_vos[name] = set([vo])

    sorted_vos = sorted(site_vos.keys(), key=lambda x: x.lower())
    site_vos = [{s: list(site_vos[s])} for s in sorted_vos]
    return HttpResponse(json.dumps(site_vos, sort_keys=True))


def tiers_report_data(request):
    """
    returns availability/reliability values for the given site and profile
    between the specified dates. The response is a collection of [date, value]
    as a json string.
    """
    error = lambda e: HttpResponse(json.dumps({'error': e}))

    group_id = request.GET.get('group_id')
    profile_id = request.GET.get('profile_id')
    if not group_id or not profile_id:
        return error('no group_id or profile_id provided')

    start = request.GET.get('start')
    end = request.GET.get('end')
    try:
        start = datetime(*(time.strptime(start, '%Y-%m-%dT%H:%M:%S')[0:6]))
        end = datetime(*(time.strptime(end, '%Y-%m-%dT%H:%M:%S')[0:6]))
    except ValueError:
        return error('invalid date format')

    values = request.GET.get('values')
    if values not in ('availability', 'reliability'):
        values = 'availability'

    avl = ACE_D_SITE_AVAILABILITY.objects.filter(
        GROUPING=group_id,
        PROFILE=profile_id,
        TIMESTAMP__gte=start,
        TIMESTAMP__lt=end
    ).order_by('TIMESTAMP')

    get_value = lambda v: v.AVAILABILITY
    if values == 'reliability':
        get_value = lambda v: v.RELIABILITY

    def norm(v):
        if v == -1:
            return None
        return v * 100.0
    data = [[v.TIMESTAMP.strftime('%b %d'), norm(get_value(v))] for v in avl]
    return HttpResponse(json.dumps(data))
