from datetime import datetime
import time
import StringIO
import csv as ocsv

from django.http import HttpResponse
from django.shortcuts import render_to_response
from django.utils import simplejson
from django.db.models import Count
from django.views.decorators.cache import cache_page

from mywlcg.mdb.models import Service
from mywlcg.mdb.models import VoServiceGroup
from mywlcg.mdb.models import Profile
from mywlcg.mdb.models import StatusChangeServiceProfileLatest
from mywlcg.mdb.models import StatusChangeServiceProfile
from mywlcg.filters.lib import parse_filters, parse_params
from mywlcg.filters.views import parse_selection, parse_common_filters
from mywlcg.base.views import path_prefix

@cache_page(600)
def index(request):
    results = {}
    results['params'] = parse_params(request)
    results = parse_selection(request, results)
    results = parse_common_filters(request, results)
    return render_to_response('services_index.html', results)

def get_results(request):
    # Initial setup of data
    profile, queryset = parse_filters(request, Service.objects)
    queryset = queryset.distinct().select_related()
    #iTotalRecords = queryset.count()
    results = []

    # Results
    for row in queryset:
        status = ''
        updated = ''
        #status_query = row.statuschangeserviceprofile_set.filter(profile = profile).values_list('metricstatus__description')
        status_query = StatusChangeServiceProfileLatest.objects.filter(\
            profile = profile, service = row.id)\
            .values('metricstatus__description', 'timestamp')
        if status_query:
            status = status_query[0]['metricstatus__description']
            timestamp = status_query[0]['timestamp']
            updated = datetime.utcfromtimestamp(timestamp).strftime("%Y-%m-%d %H:%M")

        results.append({'id': row.id,
                        'hostname': row.hostname,
                        'site_name': row.site.all()[0].sitename,
                        'flavour_name': row.flavour.flavourname,
                        'flavour_id': row.flavour.id,
                        'status': status,
                        'updated': updated,
        })

    return results

@cache_page(600)
def json(request):
    results = get_results(request)
    mydump = []
    for result in results:
        tmp = {}
        tmp['site_name'] = result['site_name']
        tmp['hostname'] = result['hostname']
        tmp['flavour_name'] = result['flavour_name']
        tmp['status'] = result['status']
        tmp['updated'] = result['updated']
        mydump.append(tmp)

    return HttpResponse(simplejson.dumps(mydump, indent=1),
        mimetype='application/javascript'
    )

@cache_page(600)
def xml(request):
    results = get_results(request)
    return render_to_response(
        'services_xml',
        mimetype = 'application/xml',
        dictionary = {
            'results': results,
        }
    )

@cache_page(600)
def csv(request):
    results = get_results(request)
    out = StringIO.StringIO()
    wrtr = ocsv.DictWriter(out,
        ["site_name", "flavour_name", "hostname", "status", "updated"],
        extrasaction='ignore',
        delimiter=','
    )
    for result in results:
        wrtr.writerow(result)
    response = HttpResponse(out.getvalue(), mimetype='text/csv')
    response['Content-Disposition'] = 'attachment; filename=services.csv'
    return response

@cache_page(600)
def json_datatables(request):
    # Datatables GET parameters
    # http://www.datatables.net/usage/server-side
    iDisplayStart = int(request.GET.get("iDisplayStart") or 0)
    iDisplayLength = int(request.GET.get("iDisplayLength") or 10)
    iColumns = int(request.GET.get("iColumns") or -1)
    sSearch = request.GET.get("sSearch")
    bEscapeRegex = request.GET.get("bEscapeRegex")
    iSortingCols = int(request.GET.get("iSortingCols") or -1)
    sEcho = int(request.GET.get("sEcho") or -1)

    # Initial setup of data
    profile, queryset = parse_filters(request, Service.objects)
    queryset = queryset.distinct().select_related()
    iTotalRecords = queryset.count()

    # Search
    if (sSearch):
        queryset = queryset.filter(hostname__icontains = sSearch)

    # Column sorting
    sorting = []
    for i in range(0, iSortingCols):
        # Parameters
        col = int(request.GET.get("iSortCol_" + str(i)) or -1)
        dir = request.GET.get("sSortDir_" + str(i))
        # Direction (asc / desc)
        filter = ''
        if dir == 'desc':
            filter = '-'
        # Column
        col_filter = ['site__sitename', 'flavour__flavourname', 'hostname']
        filter += col_filter[col]
        # Add this column sorting if needed
        if (col != -1 and filter != '' and filter != '-'):
            sorting.append(filter)
    # Make the change
    if len(sorting) > 0:
        queryset = queryset.order_by(*sorting)

    # Extracting results
    iTotalDisplayRecords = queryset.count()
    if iDisplayLength and iDisplayLength != -1:
        queryset = queryset[iDisplayStart:iDisplayStart+iDisplayLength]

    results = []
    yesterday = int(time.mktime(datetime.now().timetuple())) - 60*60*24

    for row in queryset:
        status = ''
        updated = ''
        #status_query = row.statuschangeserviceprofile_set.filter(profile = profile).values_list('metricstatus__description')
        status_query = StatusChangeServiceProfileLatest.objects.filter(\
            profile = profile, service = row.id)\
            .values('metricstatus__description', 'timestamp')
        if status_query:
            status = status_query[0]['metricstatus__description']
            timestamp = status_query[0]['timestamp']
            if timestamp < yesterday:
                #updated = '+24'
                updated = datetime.utcfromtimestamp(timestamp)
        ldap = False
        if row.flavour.flavourname in ['Site-BDII', 'Top-BDII']:
            ldap = True

        results.append({'id': row.id,
                        'hostname': row.hostname,
                        'site_name': row.site.all()[0].sitename,
                        'flavour_name': row.flavour.flavourname,
                        'flavour_id': row.flavour.id,
                        'status': status,
                        'updated': updated,
                        'ldap': ldap,
        })

    # Render
    data = {
        'iTotalRecords': iTotalRecords,
        'iTotalDisplayRecords': iTotalDisplayRecords,
        'sEcho': sEcho,
        'results': results,
        'profile': str(profile) ,
        'monitored': request.GET.get("monitored") or "1",
    }
    path_prefix(request,data)
    return render_to_response('services_json', data)

@cache_page(600)
def status_summary(request):
    # Initial setup of data
    profile, queryset = parse_filters(request, Service.objects)
    queryset = queryset.distinct()
    iTotalDisplayRecords = queryset.count()
    services = queryset.values_list('id', flat=True)
    #services = map(lambda x: x, services) # To overcome a bug in Django
    statuses = StatusChangeServiceProfileLatest.objects\
        .filter(profile = profile,  service__in = services)\
        .values('metricstatus__description').annotate(Count('metricstatus'))

    # Render
    results = {
        'statuses': statuses,
        'iTotalDisplayRecords': iTotalDisplayRecords,
    } 
    path_prefix(request,results)
    return render_to_response('services_status_summary.html', results)

@cache_page(600)
def selected(request, site_id, profile_id, unixtime):
    # Parameters
    currentdatetime = datetime.utcfromtimestamp(int(unixtime)).strftime('%x %X UTC')
    results={
		'site':site_id,
		'profile':profile_id,
                'profilename':Profile.objects.get(id=profile_id,isdeleted='N').name,
		'unixtime':unixtime,
		'currentdatetime':currentdatetime
	}
    path_prefix(request,results)
    return render_to_response('services_selected.html', results)

def selected_json(request, site_id, profile_id, unixtime):
    # Setup
    profile, queryset = parse_filters(request, Service.objects)
    sEcho = int(request.GET.get("sEcho") or -1)
    services = VoServiceGroup.objects.filter(groups=site_id).values_list('service',flat=True)
    service = queryset.filter(id__in = services).distinct()
    iTotalRecords = iTotalDisplayRecords = service.count()
    results=[]
    # Get results
    for row in service:
        status = ''
        updated = ''
        status_query = StatusChangeServiceProfile.objects.filter(\
            profile = profile_id, service = row.id, timestamp__lte = unixtime)\
            .order_by('-timestamp').values('metricstatus__description', 'timestamp')
        if status_query:
            status = status_query[0]['metricstatus__description']
            timestamp = status_query[0]['timestamp']
            updated = datetime.utcfromtimestamp(float(unixtime)).strftime("%Y-%m-%d %H:%M")
        ldap = False
        if row.flavour.flavourname in ['Site-BDII', 'Top-BDII']:
            ldap = True
        results.append({'id': row.id,
                        'hostname': row.hostname,
                        'site_name': row.site.all()[0].sitename,
                        'flavour_name': row.flavour.flavourname,
                        'flavour_id': row.flavour.id,
                        'status': status,
                        'updated': updated,
                        'ldap': ldap
        })
    # Render
    data =  {
        'iTotalRecords': iTotalRecords,
        'iTotalDisplayRecords': iTotalDisplayRecords,
        'sEcho': sEcho,
        'results': results,
        'profile': str(profile_id) ,
        'unixtime': unixtime,
        'monitored': request.GET.get("monitored") or "1",
    }
    path_prefix(request,data)
    return render_to_response('services_json_selected', data )
