import time as ptime
import calendar
from datetime import timedelta, datetime

from django.db.models.query_utils import Q
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.shortcuts import render_to_response
from django.shortcuts import get_object_or_404
from django.views.decorators.cache import cache_page

import gd
from mywlcg import settings
from mywlcg.mdb.models import (
    Profile, Service, VoServiceGroup, Downtime,
    StatusChangeServiceProfile, StatusChangeMetricService, 
    AceFlavourStatus, AceSiteStatus
)
from mywlcg.filters.lib import get_vo_profile, get_grouptype, parse_params, parse_filters
from mywlcg.filters.views import parse_selection, parse_common_filters
from mywlcg.base.views import path_prefix


@cache_page(600)
def index(request):
    # default to last week visualization
    date_or_period = request.GET.get('dateorperiod')
    if date_or_period not in ('dt', 'pd'):
        date_or_period = 'pd'
    period = request.GET.get('period')
    if period not in ('pD', 'pW', 'pM', 'p6'):
        period = 'pW'

    now = datetime.utcnow()
    if date_or_period == 'pd':  # period visualization
        if period == 'pD':
            start = now - timedelta(days=1)
        elif period == 'pW':
            start = now - timedelta(weeks=1)
        elif period == 'pM':
            start = now - timedelta(days=31)
        elif period == 'p6':
            start = now - timedelta(days=31 * 6)
        end = now
    else:  # custom dates
        def parse_date(date, default=None):
            try:
                return datetime(*(ptime.strptime(date, '%d-%m-%Y')[0:6]))
            except ValueError:
                return default
        start = parse_date(request.GET.get('startdate'), now - timedelta(weeks=1))
        end = parse_date(request.GET.get('enddate'), now)

    results = {
        'dateorperiod': date_or_period,
        'period': period,
        'starttime': int(ptime.mktime(start.timetuple())),
        'endtime': int(ptime.mktime(end.timetuple())),
        'startdate': start.strftime('%d-%m-%Y'),
        'enddate': end.strftime('%d-%m-%Y')
    }

    results['params'] = parse_params(request)
    results = parse_selection(request, results)
    results = parse_common_filters(request, results)

    all = lambda l: False not in map(bool, l)
    show_services = all(
        (len(results['services']), not len(results['flavour']), not len(results['sites']), )
    )
    # show services
    if not settings.ACE or show_services:
        results['view'] = 0
    # show flavors
    elif len(results['flavour']) and not len(results['sites']):
        results['view'] = 1
    # show sites
    else:
        results['view'] = 2

    return render_to_response('ss_index.html', results)


@cache_page(600)
def json(request):
    # Datatables GET parameters
    # http://www.datatables.net/usage/server-side
    iDisplayStart = int(request.GET.get("iDisplayStart") or 0)
    iDisplayLength = int(request.GET.get("iDisplayLength") or 10)
    iColumns = int(request.GET.get("iColumns") or -1)
    sSearch = request.GET.get("sSearch")
    bEscapeRegex = request.GET.get("bEscapeRegex")
    iSortingCols = int(request.GET.get("iSortingCols") or -1)
    sEcho = int(request.GET.get("sEcho") or -1)

    # Initial setup
    starttime = request.GET.get("starttime") or 0
    endtime = request.GET.get("endtime") or 0
    metric = request.GET.get("metric") or 0
    view=int(request.GET.get("view") or 0)
    expand=int(request.GET.get("expand") or 0)
    revdate=datetime.utcfromtimestamp(float(endtime)).date()
    production = int(request.GET.get("production", 1))
    preproduction = int(request.GET.get("preproduction", 1))

    # Ruler
    ruler = generate_ruler(int(starttime), int(endtime))

    # Extracting results
    vos,vo,profile_id=get_vo_profile(request)
    revdate=datetime.utcfromtimestamp(float(endtime)).date()
    results={
        'profile_id': str(profile_id),
        'mark_25th': ruler['mark_25th'],
        'mark_50th': ruler['mark_50th'],
        'mark_75th': ruler['mark_75th'],
        'end_marker': ruler['end_marker'],
        'starttime': starttime,
        'endtime': endtime,
    }

    if metric:
        service = request.GET.get("service") or 0
        flav = Service.objects.get(id=service).flavour
        cquery = Profile.objects.filter(id=profile_id,isdeleted='N')[0]\
                        .metricinstances.filter(atp_service_type_flavour=flav)
        columns = cquery.values_list('metric_id','atp_vo')
        filter=Q()
        for met,vo in columns:
            filter |= Q(Q(vo = vo )|Q(vo__isnull=True),metric=met)
        metrics = StatusChangeMetricService.objects\
                                .filter(filter, service=service)\
                                .values_list('metric', 'metric__name')\
                                .order_by('metric__name').distinct()
        call='ss_json_expand'
        results['name'] = 'metric'
        results['services'] = metrics
        results['service'] = service
        results['view'] = -1
    else:
        profile_id, queryset = parse_filters(request, Service.objects)
        queryset = queryset.distinct().select_related()
        # Search
        if (sSearch):
            queryset = queryset.filter(hostname__icontains = sSearch)
        if view==0:
            services = queryset.values_list('id', 'hostname',
                 'flavour__flavourname')
            iTotalRecords = queryset.count()
            results['name'] = 'service'
        else:

            services1 = queryset.values_list('id')
            voname=Profile.objects.get(id=profile_id,isdeleted='N').atp_vo.voname
            gtype=get_grouptype(voname)
            services = VoServiceGroup.objects.filter(isdeleted = 'N', service__in = services1,
                                 groups__group_type__typename = gtype)

            if not production:
                services = services.exclude(service__site__infrasttype='Production')
            if not preproduction:
                services = services.exclude(service__site__infrasttype='PPS')

            if view==1:
                services=services.values_list('service__flavour', 'service__flavour__flavourname',
                                  'groups__groupname','groups')\
                          .order_by('groups__groupname')\
                          .distinct()
                results['name'] = 'flavour'
            else:
                services=services.values_list('groups','groups__groupname').order_by('groups__groupname')\
                          .distinct()
                results['name'] = 'site'

            iTotalRecords = services.count()

        if iDisplayLength and iDisplayLength != -1:
            services = services[iDisplayStart:iDisplayStart+iDisplayLength]
        if expand==0:
            call='ss_json'
        else:
            call='ss_json_expand'
        results['services'] = services
        results['view'] = view
        results['iTotalRecords'] = iTotalRecords
        results['iTotalDisplayRecords'] = iTotalRecords
        results['sEcho'] = sEcho
        path_prefix(request,results)
    # Return
    return render_to_response(call, results)


@cache_page(600)
def graph(request, service_id, profile_id, starttime, endtime):
    # Init variables
    metric_id=request.GET.get("metric_id") or 0
    view=int(request.GET.get("view") or 0)
    site=request.GET.get("site")
    image_width = 800
    im = gd.image((image_width, 2))
    start=starttime
    end=endtime
    color = {
         '-1' : im.colorAllocate((68, 68, 68)),    # n/a (dark grey)
         'OK' : im.colorAllocate((68, 255, 68)),   # ok (green)
         'WARNING' : im.colorAllocate((194, 255, 173)),  # warning (yellow)
         'CRITICAL' : im.colorAllocate((255, 68, 68)),   # critical (red)
         'UNKNOWN' : im.colorAllocate((136, 136, 136)), # unknown (grey)
         'MISSING' : im.colorAllocate((100, 109, 126)),   # missing (dark yellow)
         'REMOVED' : im.colorAllocate((187, 136, 255)), # removed (purple)
         '99' : im.colorAllocate((255, 255, 68)),  # downtime (yellow)
    }
    black = im.colorAllocate((0, 0, 0))
    #profile_id, fqan_id = profile_id.split('-')
    starttime = int(starttime)
    endtime = int(endtime)
    total_time = endtime - starttime
    if total_time == 0:
        im.line((0, 0), (image_width-1, 0), color['-1'])
        response = HttpResponse(mimetype="image/png")
        im.writePng(response)
        return response
    if view==-1:
        revdate=datetime.utcfromtimestamp(float(endtime)).date()
        flav = Service.objects.filter(id = service_id).values_list('flavour')[0][0]
        vo = Profile.objects.get(id=profile_id,isdeleted='N').atp_vo.id
        first_query = StatusChangeMetricService.objects.filter(
                                service = service_id,
                                metric = metric_id,
                                timestamp__lt = starttime)
        queryset = StatusChangeMetricService.objects.filter(
                                service=service_id,
                                metric = metric_id,
                                timestamp__gte = starttime,
                                timestamp__lte = endtime)
        filter = Q(vo = vo ) | Q(vo__isnull=True)
        first_query = first_query.filter(filter)
        queryset = queryset.filter(filter)
        downtimes = []
    else:
        if view==0:
            first_query = StatusChangeServiceProfile.objects.filter(
                  service=service_id, profile=profile_id,
                  timestamp__lt=starttime
            )
            queryset = StatusChangeServiceProfile.objects.filter(
                 service=service_id, profile=profile_id,
                 timestamp__gte=starttime, timestamp__lte=endtime
            )
            downtimes = Downtime.objects.filter(
                servicedowntime__service=service_id,
                classification='SCHEDULED',
                isdeleted='N'
            ).values_list('starttimestamp', 'endtimestamp')

        elif view==1:
            start=datetime.utcfromtimestamp(starttime)
            end=datetime.utcfromtimestamp(endtime)
            first_query = AceFlavourStatus.objects.filter(
                   service_type_flavour = service_id, profile = profile_id,
                   grouping__groupname = site, timestamp__lt = start)
            queryset = AceFlavourStatus.objects.filter(
                   service_type_flavour = service_id, profile = profile_id,
                   grouping__groupname = site, timestamp__gte = start,
                   timestamp__lte = end)
            downtimes = AceFlavourStatus.objects.filter(
                   service_type_flavour = service_id, profile = profile_id,
                   grouping__groupname = site, downtime_flag='1')\
                   .values_list('timestamp', 'end_timestamp')
        elif view==2:
            start=datetime.utcfromtimestamp(starttime)
            end=datetime.utcfromtimestamp(endtime)
            first_query = AceSiteStatus.objects.filter(
                   grouping = service_id, profile = profile_id,
                   timestamp__lt = start)
            queryset = AceSiteStatus.objects.filter(
                   grouping = service_id, profile = profile_id,
                   timestamp__gte = start,
                   timestamp__lte = end)
            downtimes = AceSiteStatus.objects.filter(
                   grouping = service_id, profile = profile_id,
                   downtime_flag='1')\
                   .values_list('timestamp', 'end_timestamp')
    first_query = first_query.order_by('-timestamp').values_list('metricstatus__description', 'timestamp')
    queryset = queryset.order_by('timestamp')
    if first_query:
        first_query = first_query[0]
        timestamp=first_query[1]
        status_changes = [(first_query[0], timestamp)]
    else:
        status_changes = [('-1', start )]
    status_changes.extend(queryset.values_list('metricstatus__description', 'timestamp'))
    decile_out = 0 # used to calculate the reminder area
    first = True

    # Body
    if(total_time > 0):
        for change in status_changes:
            change_time = change[1]
            if (first):
                if (change_time <= start):
                    change_time = start;
                    decile1 = 0
                else:
                    if view>0:
                        diff1 = change_time - start
                        diff = diff1.days * 86400 + diff1.seconds
                        decile1 = (int)((float)(diff) / total_time * image_width)
                    else:
                        decile1 = (int)((float)(change_time - starttime) / total_time * image_width)
                    im.line((0, 0), (decile1, 0), black)
                    im.line((0, 1), (decile1, 1), black)
                status = change[0]
                decile_out = decile1;
                first = False;
            else:
                next_status = change[0]
                if view>0:
                    diff1 = change_time - start
                    diff = diff1.days * 86400 + diff1.seconds
                    decile2 = (int)((float)(diff) / total_time * image_width)
                else:
                    decile2 = (int)((float)(change_time - starttime) / total_time * image_width)
                im.line((decile1, 0), (decile2, 0), color[status])
                im.line((decile1, 1), (decile2, 1), color[status])
                size = (decile2 - decile1)
                decile_out += size
                status = next_status
                decile1 = decile2

        if len(status_changes) > 0:
            # fill leftover
            im.line((decile_out, 0), (image_width, 0), color[status])
            im.line((decile_out, 1), (image_width, 1), color[status])
        else:
            # no data?
            im.line((0, 0), (image_width, 0), black)
            im.line((0, 1), (image_width, 1), black)

    # Downtimes
    for downtime in downtimes:
        down_starttime = int(calendar.timegm(downtime[0].timetuple()))
        down_endtime = int(calendar.timegm(downtime[1].timetuple()))
        start_p = (int)((float)(down_starttime - starttime) / total_time * image_width)
        end_p = (int)((float)(down_endtime - starttime) / total_time * image_width)
        im.line((start_p, 0), (end_p, 0), color['99'])
    # Return
    response = HttpResponse(mimetype="image/png")
    im.writePng(response)
    return response


def generate_ruler(start_time, end_time):
    period = end_time - start_time;
    if period < 3600*24:
        marker = "%H:%M"
    elif period < 3600*24*30:
        marker = "%b %d, %H:%M"
    else:
        marker = "%Y-%m-%d"

    q25 = start_time + period / 4.
    mark_25th = ptime.strftime(marker, ptime.gmtime(q25))
    q50 = start_time + period / 4. * 2
    mark_50th = ptime.strftime(marker, ptime.gmtime(q50))
    q75 = start_time + period / 4. * 3
    mark_75th = ptime.strftime(marker, ptime.gmtime(q75))
    end_marker = ptime.strftime(marker, ptime.gmtime(end_time))

    return {
        'mark_25th': mark_25th,
        'mark_50th': mark_50th,
        'mark_75th': mark_75th,
        'end_marker': end_marker,
    }


def one(request, hostname, flavour, startdate='', enddate=''):
    service = get_object_or_404(Service, hostname=hostname, flavour__flavourname=flavour)
    service_id = service.id
    if enddate != '':
        redir = HttpResponseRedirect("../../../../?filters-value-Hostnames=%s&startdate=%s&enddate=%s"\
            % (service_id, startdate, enddate,))
    elif startdate != '':
        redir = HttpResponseRedirect("../../../?filters-value-Hostnames=%s&startdate=%s"\
            % (service_id, startdate,))
    else:
        redir = HttpResponseRedirect("../../?filters-value-Hostnames=%s"\
            % (service_id,))
    return redir
