import string
from string import lower
import time, calendar
import re
import simplejson
from datetime import datetime,date,timedelta
from operator import itemgetter, setitem
from itertools import groupby

from django.db import connection, DatabaseError
from django.db.models import Count, Max
from django.db.models.query_utils import Q
from django.core.exceptions import ObjectDoesNotExist
from django.shortcuts import render_to_response, get_object_or_404
from django.http import HttpResponse, HttpResponseBadRequest, Http404, HttpResponseNotFound
from django.views.decorators.cache import cache_page

from mywlcg import settings
from mywlcg.mdb.models import *
from mywlcg import base
from mywlcg.base.views import path_prefix
from mywlcg.filters.lib import get_grouptype
from mywlcg.ws.performancelimits import throttle,error


#################################################################
# MyWLCG Web Services
# https://tomtools.cern.ch/confluence/display/SAM/MyWLCG+Web+Services+Specification
#################################################################

# Index
def index(request):
    results = {}
    path_prefix(request,results)
    return render_to_response('ws_index.html', results)

# Site Names
# /sam-pi/site_names?[infrast_name]&[infrast_type]&[is_certified]&[output]
# infrast_name: Optional but may appear more than once. Possible values are: EGEE, NDGF, OSG.
# infrast_type: Optional. Possible values are: PPS, Production, SC, Test. Default: Production.
# is_certified: Optional. Possible values are: Y, N. Default: Y.
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD)
def site_names(request):
    # Params
    infrast_name = request.GET.getlist("infrast_name")
    if len(infrast_name) == 0 or infrast_name[0] == "": infrast_name = False
    infrast_type = request.GET.get("infrast_type") or "Production"
    is_certified = request.GET.get("is_certified") or "Y"
    output = lower(request.GET.get("output") or "xml")
    # Body
    values = {}
    queryset = Site.objects.filter(
        infrasttype = infrast_type,
        certifstatus = is_certified
    )
    if infrast_name:
        queryset = queryset.filter(infrast__infrastname__in = infrast_name)
    values["sites"] = queryset.values("sitename", "infrast__infrastname")
    # Output
    if output == "json":
        return render_to_response(
            'ws_site_names_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_site_names_xml',
            values,
            mimetype = 'application/xml',
        )


# Service Types and Flavours
# /sam-pi/service_types_and_flavours?[output]
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD)
def service_types_and_flavours(request):
    # Params
    values = {}
    output = lower(request.GET.get("output") or "xml")
    # Body
    values['services'] = ServiceTypeFlavour.objects \
                            .order_by('flavourname') \
                            .values_list('flavourname')
    # Output
    if output == "json":
        return render_to_response(
            'ws_service_types_and_flavours_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_service_types_and_flavours_xml',
            values,
            mimetype = 'application/xml',
        )


# Status
# /sam-pi/status?[output]
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def status(request):
    # Params
    values = {}
    output = lower(request.GET.get("output") or "xml")
    # Body
    values['results'] = MetricStatus.objects\
        .values_list('description', 'description_nagios')
    # Output
    if output == "json":
        return render_to_response(
            'ws_status_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_status_xml',
            values,
            mimetype = 'application/xml',
        )


# Services in maintenance
# /sam-pi/service_maintenance?[host_name]&[output]
# host_name: Mandatory and may appear more than once. Eg: srm.cern.ch
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def service_maintenance(request):
    # Params
    values = {}
    host_name = request.GET.getlist("host_name")
    if len(host_name) == 0 or host_name[0] == "": host_name = False
    output = lower(request.GET.get("output") or "xml")
    values['results'] = []
    # Body
    if not host_name:
        return HttpResponseBadRequest(error(output,3))
    now = datetime.utcnow()
    services = Service.objects.filter(hostname__in = host_name)\
        .values_list('id', 'hostname', 'flavour__flavourname')
    for service in services:
        downtimes = Downtime.objects.filter(
            service = service[0],
            starttimestamp__lt = now,
            isdeleted = 'N'
        ).values_list('classification')
        if len(downtimes) > 0:
            if downtimes[0][0] == 'SCHEDULED':
                in_maintenance = 'Y'
            else:
                in_maintenance = 'N'
            values['results'].append([
                service[1],
                in_maintenance
            ])
    # Output
    if output == "json":
        return render_to_response(
            'ws_service_maintenance_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_service_maintenance_xml',
            values,
            mimetype = 'application/xml',
        )


# Service Types and Flavours running in Services or Sites
# /sam-pi/service_types_and_flavours_per_service?[host_name]&[site_name]&[infrast_name]&[output]
# host_name: Optional but may appear more than once. Eg: srm.cern.ch
# site_name: Optional but may appear more than once. Eg: CERN-PROD
# infrast_name: Optional but may appear more than once. Possible values are: EGEE, NDGF, OSG
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def service_types_and_flavours_per_service(request):
    # Params
    values = {}
    host_name = request.GET.getlist("host_name")
    if len(host_name) == 0 or host_name[0] == "": host_name = False
    site_name = request.GET.getlist("site_name")
    if len(site_name) == 0 or site_name[0] == "": site_name = False
    infrast_name = request.GET.getlist("infrast_name")
    if len(infrast_name) == 0 or infrast_name[0] == "": infrast_name = False
    output = lower(request.GET.get("output") or "xml")
    values['results'] = []
    # Body
    
    sites = Site.objects
    if site_name:
        sites = sites.filter(sitename__in = site_name)
    if infrast_name:
        sites = sites.filter(infrast__infrastname__in = infrast_name)
    sites = sites.values(
        'id',
        'sitename',
        'infrast__infrastname',
    )

    for site in sites:
        services = Service.objects.filter(site = site['id'])
        if host_name:
            services = services.filter(hostname__in = host_name)
        services = services.values_list(
                'hostname',
                'flavour__flavourname',
                'flavour__flavourname'
            )
        if len(services) > 0:
            values['results'].append([
                site['sitename'],
                site['infrast__infrastname'],
                services
            ])
    # Output
    if output == "json":
        return render_to_response(
            'ws_service_types_and_flavours_per_service_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_service_types_and_flavours_per_service_xml',
            values,
            mimetype = 'application/xml',
        )


# VO Names
# /sam-pi/vo_names?[output]
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def vo_names(request):
    # Params
    values = {}
    output = lower(request.GET.get("output") or "xml")
    # Body
    values['results'] = Vo.objects.values_list('voname', flat=True)
    # Output
    if output == "json":
        return render_to_response(
            'ws_vo_names_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_vo_names_xml',
            values,
            mimetype = 'application/xml',
        )


# VOs supported by each Service or Site
# /sam-pi/vos_per_service?[vo_name]&[site_name]&[infrast_name]&[service_type]&[service_type_flavour]&[output]
# vo_name: Optional but may appear more than once. Eg: Alice
# site_name: Optional but may appear more than once. Eg: CERN-PROD
# infrast_name: Optional but may appear more than once. Possible values are: EGEE, NDGF, OSG
# service_type: Optional but may appear more than once. Eg: LFC
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def vos_per_service(request):
    # Params
    values = {}
    vo_name = request.GET.getlist("vo_name")
    if len(vo_name) == 0 or vo_name[0] == "": vo_name = False
    site_name = request.GET.getlist("site_name")
    if len(site_name) == 0 or site_name[0] == "": site_name = False
    infrast_name = request.GET.getlist("infrast_name")
    if len(infrast_name) == 0 or infrast_name[0] == "": infrast_name = False
    service_type = request.GET.getlist("service_type")
    service_type_flavour = request.GET.getlist("service_type_flavour")
    service_type_flavour = service_type_flavour + service_type 
    output = lower(request.GET.get("output") or "xml")
    values['results'] = []
    # Body
    # It looks horrible, but Django 1.3 will improve the values method to
    # retrieve easily N to N relationships. :)
    # http://docs.djangoproject.com/en/dev/ref/models/querysets/#values-fields
    sites = Site.objects

    if site_name:
        sites = sites.filter(sitename__in = site_name)
    if infrast_name:
        sites = sites.filter(infrast__infrastname__in = infrast_name)
    if not site_name and not infrast_name:
        sites = sites.all()

    for site in sites:
        services = []
        services_objs = Service.objects.filter(site = site)
        if service_type_flavour:
            services_objs = services_objs.filter(flavour__flavourname__in = service_type_flavour)
        if vo_name:
            services_objs = services_objs.filter(vo__voname__in = vo_name)
        for service in services_objs:
            vos = Vo.objects.filter(service = service).values_list('voname', flat=True)
            services.append([
                service.hostname,
                service.flavour.flavourname,
                service.flavour,
                vos
            ])
        if len(services) > 0:
            values['results'].append([
                site.sitename,
                site.infrast.infrastname,
                services
            ])
    # Output
    if output == "json":
        return render_to_response(
            'ws_vos_per_service_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_vos_per_service_xml',
            values,
            mimetype = 'application/xml',
        )


# Groups defined by VOs
# /sam-pi/groups_defined_by_vos?[vo_name]&[group_type]&[output]
# vo_name: Optional but may appear more than once. Eg: Alice
# group_type: Optional but may appear more than once. Eg: LHCb_Site
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def groups_defined_by_vos(request):
    # Params
    values = {}
    vo_name = request.GET.getlist("vo_name")
    if len(vo_name) == 0 or vo_name[0] == "": vo_name = False
    group_type = request.GET.getlist("group_type")
    if len(group_type) == 0 or group_type[0] == "": group_type = False
    output = lower(request.GET.get("output") or "xml")
    # Body
    vos = Vo.objects
    if vo_name:
        vos = vos.filter(voname__in = vo_name)
    vos = vos.values_list('id', 'voname')

    results = []
    for vo in vos:
        groups = VoGroup.objects.filter(vo = vo[0], groups__isdeleted = 'N')
        if group_type:
            groups = groups.filter(groups__group_type__typename__in = group_type)
        groups = groups.values_list(
                'groups__groupname',
                'groups__description',
                'groups__group_type__typename',
                'groups__group_type__description',
            ).order_by('groups__group_type__typename', 'groups__groupname')
        if len(groups) > 0:
            # build index dict of gr_type_name: gr_type_name_description
            group_type_ind = dict(set(map(lambda x: (x[2], x[3]), groups)))
            
            results_types = []
            # group_by function works like UNix uniq command
            # in case a full SQL group by functionality is needed the code
            # below will not work
            for gr_type_key, gr_items in groupby(groups, itemgetter(2)):
                results_types.append([gr_type_key, 
                                      group_type_ind[gr_type_key],   
                                      list(gr_items)])
                
            results.append([vo[1], results_types])
    values['results'] = results
    # Output
    if output == "json":
        return render_to_response(
            'ws_groups_defined_by_vos_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_groups_defined_by_vos_xml',
            values,
            mimetype = 'application/xml',
        )


# Services defined in groups
# /sam-pi/services_defined_in_groups?[vo_name]&[group_name]&[group_type]&[output]
# vo_name: Mandatory and may appear more than once. Eg: lhcb
# group_name: Optional and may appear more than once. Eg: LCG.BHAM-HEP.uk
# group_type: Optional and may appear more than once. Eg: LHCb_Site
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def services_defined_in_groups(request):
    # Params
    values = {}
    vo_name = request.GET.getlist("vo_name")
    if len(vo_name) == 0 or vo_name[0] == "": vo_name = False
    group_name = request.GET.getlist("group_name")
    if len(group_name) == 0 or group_name[0] == "": group_name = False
    group_type = request.GET.getlist("group_type")
    if len(group_type) == 0 or group_type[0] == "": group_type = False
    output = lower(request.GET.get("output") or "xml")
    # Body
    if not vo_name:
        return HttpResponseBadRequest(error(output,3))

    qs = VoServiceGroup.objects.filter(vo__voname__in = vo_name,isdeleted='N')
    if group_name:
        qs = qs.filter(groups__groupname__in = group_name)
    if group_type:
        qs = qs.filter(groups__group_type__typename__in = group_type)
    qs = qs.values(
        'vo__voname',
        'groups__group_type__typename',
        'groups__group_type__description',
        'groups__groupname',
        'groups__description',
        'service__hostname',
        'service__flavour__flavourname',
        'service__flavour__flavourname'
    )
    results = {}
    for vsg in qs:
        vo = vsg['vo__voname']
        group_type = vsg['groups__group_type__typename']
        group = vsg['groups__groupname']
        if vo not in results:
            results[vo] = {
                "group_types": {}
            }
        if group_type not in results[vo]["group_types"]:
            results[vo]["group_types"][group_type] = {
                "groups": {},
                "description": vsg['groups__group_type__description']
            }
        if group not in results[vo]["group_types"][group_type]["groups"]:
            results[vo]["group_types"][group_type]["groups"][group] = {
                "services": [],
                "description": vsg['groups__description']
            }
        results[vo]["group_types"][group_type]["groups"][group]["services"].append({
            "hostname": vsg['service__hostname'],
            "flavour": vsg['service__flavour__flavourname'],
            "type": vsg['service__flavour__flavourname']
        })

    # Output
    values = {}
    values['results'] = results
    if output == "json":
        return render_to_response(
            'ws_services_defined_in_groups_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_services_defined_in_groups_xml',
            values,
            mimetype = 'application/xml',
        )


# Service types & flavours defined in groups
# /sam-pi/service_types_and_flavours_defined_in_groups?[vo_name]&[group_name]&[group_type]&[output]
# vo_name: Mandatory and may appear more than once. Eg: lhcb
# group_name: Optional and may appear more than once. Eg: LCG.BHAM-HEP.uk
# group_type: Optional and may appear more than once. Eg: LHCb_Site
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def service_types_and_flavours_defined_in_groups(request):
    # Params
    values = {}
    vo_name = request.GET.getlist("vo_name")
    if len(vo_name) == 0 or vo_name[0] == "": vo_name = False
    group_name = request.GET.getlist("group_name")
    if len(group_name) == 0 or group_name[0] == "": group_name = False
    group_type = request.GET.getlist("group_type")
    if len(group_type) == 0 or group_type[0] == "": group_type = False
    output = lower(request.GET.get("output") or "xml")
    # Body
    if not vo_name:
        return HttpResponseBadRequest(error(output,3))

    qs = VoServiceGroup.objects.filter(vo__voname__in = vo_name)
    if group_name:
        qs = qs.filter(groups__groupname__in = group_name)
    if group_type:
        qs = qs.filter(groups__group_type__typename__in = group_type)
    qs = qs.distinct().values(
        'vo__voname',
        'groups__group_type__typename',
        'groups__group_type__description',
        'groups__groupname',
        'groups__description',
        'service__flavour__flavourname',
        'service__flavour__flavourname'
    )
    results = {}
    for vsg in qs:
        vo = vsg['vo__voname']
        group_type = vsg['groups__group_type__typename']
        group = vsg['groups__groupname']
        if vo not in results:
            results[vo] = {
                "group_types": {}
            }
        if group_type not in results[vo]["group_types"]:
            results[vo]["group_types"][group_type] = {
                "groups": {},
                "description": vsg['groups__group_type__description']
            }
        if group not in results[vo]["group_types"][group_type]["groups"]:
            results[vo]["group_types"][group_type]["groups"][group] = {
                "services": [],
                "description": vsg['groups__description']
            }
        results[vo]["group_types"][group_type]["groups"][group]["services"].append({
            "flavour": vsg['service__flavour__flavourname'],
            "type": vsg['service__flavour__flavourname']
        })

    # Output
    values = {}
    values['results'] = results
    if output == "json":
        return render_to_response(
            'ws_service_types_and_flavours_defined_in_groups_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_service_types_and_flavours_defined_in_groups_xml',
            values,
            mimetype = 'application/xml',
        )


# Metrics defined in profiles
# /sam-pi/metrics_in_profiles?[vo_name]&[namespace]&[profile_name]&[output]
# vo_name: Mandatory and may appear more than once. Eg: lhcb
# namespace: Optional and may appear more than once. Eg: ch.cern.sam
# profile_name: Optional and may appear more than once. Eg: ALICE_CRITICAL
# service_type_flavour: Optional and may appear more than once. Eg: CREAM-CE
# service_type: Optional and may appear more than once. Eg: CE
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def metrics_in_profiles(request):
    # Params
    vo_name = request.GET.getlist("vo_name")
    if len(vo_name) == 0 or vo_name[0] == "": vo_name = False
    namespace = request.GET.getlist("namespace")
    if len(namespace) == 0 or namespace[0] == "": namespace = False
    profile_name = request.GET.getlist("profile_name")
    if len(profile_name) == 0 or profile_name[0] == "": profile_name = False
    service_type_flavour = request.GET.getlist("service_type_flavour")
    if len(service_type_flavour) == 0 or service_type_flavour[0] == "": service_type_flavour = False
    service_type = request.GET.getlist("service_type")
    if len(service_type) == 0 or service_type[0] == "": service_type = False
    if service_type and not service_type_flavour:
       service_type_flavour = service_type
    elif service_type and service_type_flavour:
       service_type_flavour = service_type_flavour + service_type
       
    output = lower(request.GET.get("output") or "xml")
    # Body
    if not vo_name:
        return HttpResponseBadRequest(error(output,3))
    profiles = Profile.objects.filter(atp_vo__voname__in = vo_name,isdeleted='N')
    if namespace:
        profiles = profiles.filter(namespace__name__in = namespace)
    if profile_name:
        profiles = profiles.filter(name__in = profile_name)
    profiles=profiles.values_list('name','namespace__name','atp_vo__voname','version','valid_from','tags','description')\
	.order_by(
            'atp_vo__voname',
            'name',
        )
    values = {}
    results = {}
    for profile,namespace,voname,version,valid_from,tags,description in profiles:
        pmm=Profile.objects.filter(name=profile,namespace__name=namespace,isdeleted='N')[0].metricinstances
        if service_type_flavour:
            pmm = pmm.filter(atp_service_type_flavour__flavourname__in = service_type_flavour)
        pmm = pmm.values_list(
                                'metric__name',
                                'atp_service_type_flavour__flavourname',
                                'fqan__fqan'
                )
        if voname not in results:
            results[voname] = {}
        if profile not in results[voname]:
            results[voname][profile] = {'tags':tags,'description':description,'namespaces':{}}
        if namespace not in results[voname][profile]['namespaces']:
            results[voname][profile]['namespaces'][namespace] = []
        for pm in pmm:
           results[voname][profile]['namespaces'][namespace].append([
               pm[0],version,pm[0],valid_from,valid_from,'None',pm[1],pm[1],voname,pm[2]
            ])

    values['results'] = results
    # Output
    if output == "json":
        return render_to_response(
            'ws_metrics_in_profiles_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_metrics_in_profiles_xml',
            values,
            mimetype = 'application/xml',
        )


# History of metrics in a profile
# /sam-pi/metrics_history_in_profile?[vo_name]&[namespace]&[profile_name]&[start_time]&[end_time]&[service_flavour]&[service_hostname]&[output]
# vo_name: Mandatory and may appear more than once. Eg: lhcb
# namespace: Optional and may appear more than once. Eg: ch.cern.sam
# profile_name: Mandatory and may appear more than once. Eg: ALICE_CRITICAL
# start_time: Mandatory. UTC time in W3C format: 'YYYY-MM-DDThh:mm:ssZ'. Eg: 2010-10-01T13:22:15Z
# end_time: Mandatory. UTC time in W3C format: 'YYYY-MM-DDThh:mm:ssZ'. Eg: 2010-10-02T00:00:00Z
# service_flavour: Optional and may appear more than once. Eg: CREAM-CE
# service_hostname: Optional and may appear more than once. Eg: alice16.spbu.ru
# metric_name: Optional and may appear more than once. Eg: hr.srce.CADist-Check 
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def metrics_history_in_profile(request):
    # Params
    vo_name = request.GET.getlist("vo_name")
    if len(vo_name) == 0 or vo_name[0] == "": vo_name = False
    namespace = request.GET.getlist("namespace")
    if len(namespace) == 0 or namespace[0] == "": namespace = False
    profile_names = request.GET.getlist("profile_name")
    if len(profile_names) == 0 or profile_names[0] == "": profile_names = False
    start_time = request.GET.get("start_time")
    end_time = request.GET.get("end_time")
    service_flavour = request.GET.getlist("service_flavour")
    if len(service_flavour) == 0 or service_flavour[0] == "": service_flavour = False
    service_hostname = request.GET.getlist("service_hostname")
    if len(service_hostname) == 0 or service_hostname[0] == "": service_hostname = False
    metric_name = request.GET.getlist("metric_name")
    if len(metric_name) == 0 or metric_name[0] == "": metric_name = False
    output = lower(request.GET.get("output") or "xml")

    # Body
    if not vo_name or not profile_names or not start_time or not end_time:
        return HttpResponseBadRequest(error(output,3))
    try:
       start_time_e = calendar.timegm(time.strptime(start_time, '%Y-%m-%dT%H:%M:%SZ'))
       end_time_e = calendar.timegm(time.strptime(end_time, '%Y-%m-%dT%H:%M:%SZ'))
    except ValueError, (ve):
       return HttpResponseBadRequest(error(output,"400: Bad Request. " + str(ve)))

    profiles = Profile.objects.filter(
        name__in=profile_names,
        atp_vo__voname__in=vo_name,
        isdeleted='N'
    )
    if namespace:
        profiles = profiles.filter(namespace__name__in=namespace)
    profiles = profiles.values_list(
        'name', 'namespace__name', 'atp_vo__voname'
    ).order_by('atp_vo__voname','name')

    results = []
    for profile, namespace, vo in profiles:
        metricdata = MetricHistoryInProfile.objects.filter(
            profile_name=str(profile),
            namespace=str(namespace),
            metric_exec_time__gte=start_time_e,
            metric_exec_time__lte=end_time_e
        )

        if service_hostname:
            metricdata = metricdata.filter(hostname__in=service_hostname)
        if service_flavour:
            metricdata = metricdata.filter(service_flavour__in=service_flavour)
        if metric_name:
            metricdata = metricdata.filter(metric_name__in=metric_name)

        metricdata = metricdata.values_list(
                 'hostname',
                 'service_flavour',
                 'service_flavour',
                 'metric_exec_time',
                 'metric_name',
                 'metric_status',
                 'metric_output_summary',
                 'metric_result_insert_time',
                 'fqan',
                 'gathered_at'
         ).order_by('hostname', 'service_flavour', 'service_flavour', 'metric_exec_time')

        hist_metrics = []

        # preprocess
        md_list = [list(x) for x in metricdata]
        for item in md_list:
            item[3] = datetime.utcfromtimestamp(item[3]).isoformat()+'Z'
            item[7] = datetime.utcfromtimestamp(item[7]).isoformat()+'Z'
        # extract groupby fields
        prep_md = [ ((x[0:3]), x[3:]) for x in md_list ] 

        # groupby
        for el, li in groupby(prep_md, itemgetter(0)):
            # [ host, serv_type, serv_fl, [ [ metric_entry1], [metric_entry2], ... ]]
            hist_metrics.append([ s for s in el ]+[[ m[1] for m in li ]])

        # set profile header
        results.append([profile, namespace, vo, hist_metrics])

    # Output
    if output == "json":
        	return render_to_response(
            		'ws_metrics_history_in_profile_json',
            		{'results': results},
            		mimetype = 'application/json',
        	)
    else:
        	return render_to_response(
            		'ws_metrics_history_in_profile_xml',
            		{'results': results},
            		mimetype = 'application/xml',
        	)        


# /sam-pi/metric_output_details?[vo_name]&[metric_name]&[service_flavour]&[service_hostname]&[check_time]&[output]
# vo_name: Mandatory. Eg: alice
# metric_name: Mandatory. Eg: org.sam.CE-JobSubmit
# service_flavour: Mandatory. Eg: CE
# service_hostname: Mandatory. Eg: alice19.spbu.ru
# check_time: Mandatory. UTC time in W3C format: 'YYYY-MM-DDThh:mm:ssZ'. Eg: 2010-11-29T00:23:18
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def metric_output_details(request):
    # Params
    values = {}
    vo_name = request.GET.get("vo_name")
    metric_name = request.GET.get("metric_name")
    service_flavour = request.GET.get("service_flavour")
    service_hostname = request.GET.get("service_hostname")
    check_time = request.GET.get("check_time")
    output = lower(request.GET.get("output") or "xml")

    # Body
    if not vo_name or not service_flavour\
            or not service_hostname or not check_time:
        return HttpResponseBadRequest(error(output,3))
    try:
        check_time_e = calendar.timegm(time.strptime(check_time, '%Y-%m-%dT%H:%M:%SZ'))
    except ValueError, (ve):
        return HttpResponseBadRequest(error(output,"400: Bad Request. " + str(ve)))

    metricdata = MetricData.objects.select_related(depth=1).filter(
        Q(vo__isnull=True) | Q(vo__voname=vo_name),
        metric__name=metric_name,
        service__flavour__flavourname=service_flavour,
        service__hostname=service_hostname,
        check_time=check_time_e
    )

    values = {'display': 0}
    if metricdata:
        metricdata = metricdata[0]
        values = {
            'hostname': service_hostname,
            'type': service_flavour,
            'flavour': service_flavour,
            'check_time': check_time,
            'metric_name': metric_name,
            'vo__voname': vo_name,
            'fqan__name': metricdata.fqan.name,
            'metricstatus__description': metricdata.metricstatus.description,
            'summarydata': metricdata.summarydata,
            'metricdetail__detail': metricdata.metricdetail.detail or '',
            'display': 1
        }

    # Output
    if output == "json":
        return render_to_response(
            'ws_metric_output_details_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_metric_output_details_xml',
            values,
            mimetype = 'application/xml',
        )


# Latest metric results in a profile
# /sam-pi/latest_metric_results_in_profile?[vo_name]&[namespace]&[profile_name]&[service_flavour]&[service_hostname]&[group_type]&[group_name]&[output]
# vo_name: Mandatory and may appear more than once. Eg: lhcb
# namespace: Optional and may appear more than once. Eg: ch.cern.sam
# profile_name: Mandatory and may appear more than once. Eg: ALICE_CRITICAL
# service_flavour: Optional and may appear more than once. Eg: CREAM-CE
# service_hostname: Optional and may appear more than once. Eg: alice16.spbu.ru
# group_type: Optional and may appear more than once. Eg: Site
# group_name: Optional and may appear more than once. Eg:
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def latest_metric_results_in_profile(request):
    # Params
    vo_name = request.GET.getlist("vo_name")
    if len(vo_name) == 0 or vo_name[0] == "": vo_name = False
    namespace = request.GET.getlist("namespace")
    if len(namespace) == 0 or namespace[0] == "": namespace = False
    profile_name = request.GET.getlist("profile_name")
    if len(profile_name) == 0 or profile_name[0] == "": profile_name = False
    service_flavour = request.GET.getlist("service_flavour")
    if len(service_flavour) == 0 or service_flavour[0] == "": service_flavour = False
    service_hostname = request.GET.getlist("service_hostname")
    if len(service_hostname) == 0 or service_hostname[0] == "": service_hostname = False
    group_type = request.GET.getlist("group_type")
    if len(group_type) == 0 or group_type[0] == "": group_type = False
    group_name = request.GET.getlist("group_name")
    if len(group_name) == 0 or group_name[0] == "": group_name = False
    output = lower(request.GET.get("output") or "xml")

    # Body
    if not vo_name or not profile_name:
        return HttpResponseBadRequest(error(output,3))

    results = {}
    nrsults = 0

    profiles = Profile.objects.filter(
        name__in=profile_name,
        atp_vo__voname__in=vo_name,
        isdeleted='N'
    )
    if namespace:
        profiles = profiles.filter(namespace__name__in=namespace)
    profiles = profiles.values_list(
        'name','namespace__name','atp_vo__voname'
    ).order_by('atp_vo__voname','name')

    for profile, namespace, voname in profiles:
        if not profile in results:
           results[profile]={}

        results[profile][namespace] = {
            'vo': voname,
            'group_name':{}
        }
        servstatus = StatusChangeServiceProfileLatest.objects.filter(
            profile__name=profile, profile__namespace__name=namespace
        )
        metservflav = Profile.objects.filter(
            name=profile, namespace__name=namespace, isdeleted='N'
        )[0].metricinstances

        if service_flavour:
            metservflav = metservflav.filter(
                atp_service_type_flavour__flavourname__in=service_flavour
           )
        metservflav = metservflav.values_list(
            'metric','atp_service_type_flavour','atp_vo'
        ).order_by('metric','atp_service_type_flavour').distinct()

        if metservflav.count():
            services1 = VoServiceGroup.objects.filter(vo__voname=voname, isdeleted='N')
            metricdata = MetricDataLatest.objects.select_related(depth=1)
            q_obj = Q()
            for m, sf, vo in metservflav:
                q_obj |= Q(
                    Q(vo__isnull=True) | Q(vo=vo),
                    metric=m, service__flavour=sf
                )
            metricdata = metricdata.filter(q_obj)

            if service_hostname:
                metricdata = metricdata.filter(service__hostname__in=service_hostname)

            if not group_type:
                group_type = [get_grouptype(voname), ]
            for gtype in group_type:
                services = services1.filter(groups__group_type__typename=gtype)
                if not group_name:
                    group_name = services.values_list(
                        'groups__groupname', flat=True
                    ).order_by('groups__groupname').distinct('groups__groupname')
                for gname in group_name:
                    services2 = services.filter(groups__groupname=gname)
                    services2 = services2.distinct('id').values_list('service__id', flat=True)
                    metricdata2 = metricdata.filter(service__id__in=services2)

                    nrsults += metricdata2.count()
                    if nrsults > settings.DB_LIMIT:
                        return HttpResponseBadRequest(error(output, 1))

                    if not gname in results[profile][namespace]['group_name']:
                        results[profile][namespace]['group_name'][gname] = {}
                    if not gtype in results[profile][namespace]['group_name'][gname]:
                        results[profile][namespace]['group_name'][gname][gtype] = {'services': {}, }
                    results[profile][namespace]['group_name'][gname][gtype] = insert(
                        metricdata2,
                        results[profile][namespace]['group_name'][gname][gtype],
                        servstatus
                    )
    # Output
    if output == "json":
        return render_to_response(
            'ws_latest_metric_results_in_profile_json',
            {'results': results},
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_latest_metric_results_in_profile_xml',
            {'results': results},
            mimetype = 'application/xml',
        )

def insert(metricdata,res,servstatus):
        metricdata = metricdata.values(
            'service__id',
            'service__hostname',
            'service__flavour__flavourname',
            'service__flavour__flavourname',
            'check_time',
            'metric__name',
            'metricstatus__description',
            'summarydata',
            'insert_time',
            'fqan',
            'gatheredat'
        )

        for datum in metricdata:
            if datum['service__id'] not in res['services']:
                status = servstatus.filter(service=datum['service__id'])
                if not status:
                    continue
                res['services'][datum['service__id']] = {
                    'hostname': datum['service__hostname'],
                    'type': datum['service__flavour__flavourname'],
                    'flavour': datum['service__flavour__flavourname'],
                    'status': status.values_list('metricstatus__description')[0][0],
                    'metrics': {}
                }
            if datum['metric__name'] not in res['services'][datum['service__id']]['metrics'] or\
                    res['services'][datum['service__id']]['metrics'][datum['metric__name']]['insert_time'] < datum['insert_time']:
                if datum['fqan']:
                   datum['fqan']=Fqan.objects.filter(id=datum['fqan']).values_list('name')[0][0]
                if datum['gatheredat']:
                   datum['gatheredat']=Gatheredat.objects.filter(id=datum['gatheredat']).values_list('name')[0][0]
                res['services'][datum['service__id']]['metrics'][datum['metric__name']] = {
                    'check_time': datetime.utcfromtimestamp(datum['check_time']).isoformat() + 'Z',
                    'status': datum['metricstatus__description'],
                    'summarydata': datum['summarydata'],
                    'insert_time': datetime.utcfromtimestamp(datum['insert_time']).isoformat() + 'Z',
                    'fqan': datum['fqan'],
                    'gatheredat': datum['gatheredat']
                }
        return res


# Status of Services in a Profile
# /sam-pi/status_of_service_in_profile?[vo_name]&[namespace]&[profile_name]&[service_type_flavour]&[service_type]&[output]
# vo_name: Mandatory. Eg: lhcb
# namespace: Optional and may appear more than once. Eg: ch.cern.sam
# profile_name: Mandatory. Eg: ROC_CRITICAL, ALICE_CRITICAL, CMS_CRITICAL
# service_type_flavour: Optional and may appear more than once. Eg: CREAM-CE
# service_type: Optional and may appear more than once. Eg: CE
# isdeleted (Default value: 'N', Optional, Possible values: N,Y)
# isinproduction (Default value: 'Y', Optional, Possible values: N,Y)
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def status_of_service_in_profile(request):
    # Params
    vo_name = request.GET.getlist("vo_name")
    if len(vo_name) == 0 or vo_name[0] == "": vo_name = False
    namespace = request.GET.getlist("namespace")
    if len(namespace) == 0 or namespace[0] == "": namespace = False
    profile_name = request.GET.getlist("profile_name")
    if len(profile_name) == 0 or profile_name[0] == "": profile_name = False
    service_type_flavour = request.GET.getlist("service_type_flavour")
    if len(service_type_flavour) == 0 or service_type_flavour[0] == "": service_type_flavour = False
    service_type = request.GET.getlist("service_type")
    if len(service_type) == 0 or service_type[0] == "": service_type = False
    if service_type and not service_type_flavour:
       service_type_flavour = service_type
    elif service_type and service_type_flavour:
       service_type_flavour = service_type_flavour + service_type
    output = lower(request.GET.get("output") or "xml")
    isdeleted = request.GET.get("isdeleted") or "N"
    isinproduction = request.GET.get("isinproduction") or "Y"

    # Body
    if not vo_name or not profile_name or not settings.ACE:
        return HttpResponseBadRequest(error(output,3))
    results = {}
    profiles = Profile.objects.filter(name__in=profile_name,isdeleted='N')

    if namespace:
        profiles = profiles.filter(namespace__name__in=namespace)

    profiles = profiles.values_list('name','namespace__name','atp_vo__voname')\
                        .order_by('atp_vo__voname','name')
    for profile,namespace,vo in profiles:
        if not profile in results:
           results[profile]={}

        results[profile][namespace] = { 'vo': vo }
        services = AceServiceStatusLatest.objects\
				.filter(
					profile__name = profile,
                                        profile__namespace__name = namespace,
					service__isdeleted = isdeleted,
					service__isinproduction = isinproduction
				)
        if service_type_flavour:
            services = services.filter(service__flavour__flavourname__in = service_type_flavour)
        services = services.values(
            'service__id',
            'service__hostname',
            'service__flavour__flavourname',
            'service__flavour__flavourname',
            'metricstatus__description',
            'start_timestamp',
            'end_timestamp',
        )
        results[profile][namespace]['services'] = services

    # Output
    if output == "json":
        return render_to_response(
            'ws_status_of_service_in_profile_json',
            {'results': results},
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_status_of_service_in_profile_xml',
            {'results': results},
            mimetype = 'application/xml',
        )


# Status of Sites in a Profile
# /sam-pi/sites_status?[vo_name]&[namespace]&[profile_name]&[group_name]&[group_type]&[output]
# vo_name: Mandatory. Eg: lhcb
# namespace: Optional and may appear more than once. Eg: ch.cern.sam
# profile_name: Mandatory. Eg: ROC_CRITICAL, ALICE_CRITICAL, CMS_CRITICAL
# group_name: Optional and may appear more than once. Eg: AEGIS01-IPB-SCL
# group_type: Mandatory. Eg: Site
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def status_of_site_in_profile(request):
    # Params
    vo_name = request.GET.getlist("vo_name")
    if len(vo_name) == 0 or vo_name[0] == "": vo_name = False
    namespace = request.GET.getlist("namespace")
    if len(namespace) == 0 or namespace[0] == "": namespace = False
    profile_name = request.GET.getlist("profile_name")
    if len(profile_name) == 0 or profile_name[0] == "": profile_name = False
    group_name = request.GET.getlist("group_name")
    if len(group_name) == 0 or group_name[0] == "": group_name = False
    group_type = request.GET.get("group_type")
    output = lower(request.GET.get("output") or "xml")
    # Body
    if not vo_name or not profile_name or not group_type or not settings.ACE:
        return HttpResponseBadRequest(error(output,3))
    profiles = Profile.objects.filter(name__in=profile_name,isdeleted='N')

    if namespace:
        profiles = profiles.filter(namespace__name__in=namespace)

    profiles = profiles.values_list('name','namespace__name','atp_vo__voname')\
                        .order_by('atp_vo__voname','name')
    results = {}
    for profile,namespace,vo in profiles:
        if not profile in results:
           results[profile]={}

        results[profile][namespace] = { 'vo': vo }
        sites = AceSiteStatusLatest.objects.filter(
				profile__name = profile, 
				profile__namespace__name = namespace,
				grouping__group_type__typename = group_type)
        if group_name:
            sites = sites.filter(grouping__groupname__in = group_name)
        sites = sites.values(
            'grouping__id',
            'grouping__groupname',
            'metricstatus__description',
            'start_timestamp',
            'end_timestamp',
        )

        results[profile][namespace]['sites'] = sites

    # Output
    if output == "json":
        return render_to_response(
            'ws_status_of_site_in_profile_json',
            {'results': results},
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_status_of_site_in_profile_xml',
            {'results': results},
            mimetype = 'application/xml',
        )


# Downtimes
# /sam-pi/downtimes?[start_time]&[end_time]&[output]
# start_time: Mandatory. UTC time in W3C format: 'YYYY-MM-DDThh:mm:ssZ'. Eg: 2010-10-01T13:22:15Z
# end_time: Mandatory. UTC time in W3C format: 'YYYY-MM-DDThh:mm:ssZ'. Eg: 2010-10-02T00:00:00Z
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def downtimes(request):
    # Params
    start_time = request.GET.get("start_time")
    end_time = request.GET.get("end_time")
    output = lower(request.GET.get("output") or "xml")
    # Body
    if not start_time or not end_time:
        return HttpResponseBadRequest(error(output,3))
    results = []
    format = '%Y-%m-%dT%H:%M:%SZ'
    try:
       start_time_e = datetime(*time.strptime(start_time, format)[0:5]).strftime("%Y-%m-%d %H:%M:%S")
       end_time_e = datetime(*time.strptime(end_time, format)[0:5]).strftime("%Y-%m-%d %H:%M:%S")
    except ValueError, (ve):
       return HttpResponseBadRequest(error(output,"400: Bad Request. " + str(ve)))
    downtimes = Downtime.objects.filter(isdeleted = 'N').filter(
        Q(starttimestamp__lte = end_time_e, endtimestamp__gte = start_time_e) |
        Q(endtimestamp__gte = start_time_e, starttimestamp__lte = end_time_e)
    ).values(
        'id',
        'starttimestamp',
        'endtimestamp',
        'inserttimestamp',
        'classification',
        'gocdowntimeid',
        'osgdowntimeid',
        'severity',
        'description',
    )
    try:
      for downtime in downtimes:
        services = Downtime.objects.get(id = downtime['id']).service.all().values(
            'id',
            'hostname',
            'flavour__flavourname',
            'flavour__flavourname',
        )
        services_in_sites = []
        for site in Downtime.objects.get(id = downtime['id']).site.all():
            services_in_sites.extend(site.service_set.all().values(
                'id',
                'hostname',
                'flavour__flavourname',
                'flavour__flavourname',
            ))
        downtime['services'] = services
        downtime['services_in_sites'] = services_in_sites
        results.append(downtime)
    except ObjectDoesNotExist, (dne):
            return HttpResponseNotFound(error(output,"404: Page Not Found. " + str(dne)))
    # Output
    if output == "json":
        return render_to_response(
            'ws_downtimes_json',
            {'results': results},
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_downtimes_xml',
            {'results': results},
            mimetype = 'application/xml',
        )


# Service Availability in Profile
# /sam-pi/service_availability_in_profile?[vo_name]&[namespace]&[profile_name]&[start_time]&[end_time]&[type]&[service_flavour]&[service_hostname]&[output]
# vo_name: Mandatory and may appear more than once. Eg: cms
# namespace: Optional and may appear more than once. Eg: ch.cern.sam
# profile_name: Mandatory and may appear more than once. Eg: CMS_CRITICAL
# start_time: Mandatory. UTC time in W3C format: 'YYYY-MM-DDThh:mm:ssZ'. Eg: 2010-10-01T13:22:15Z
# end_time: Mandatory. UTC time in W3C format: 'YYYY-MM-DDThh:mm:ssZ'. Eg: 2010-10-02T00:00:00Z
# type: Mandatory. Possible values: 'HOURLY', 'DAILY', 'WEEKLY'.
# group_type: Optional and may appear more than once. Eg: Site
# group_name: Optional and may appear more than once. Eg:
# service_flavour: Optional and may appear more than once. Eg: SRMv2
# service_hostname: Optional and may appear more than once. Eg:
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def service_availability_in_profile(request):
    # Params
    vo_name = request.GET.getlist("vo_name")
    if len(vo_name) == 0 or vo_name[0] == "": vo_name = False
    namespace = request.GET.getlist("namespace")
    if len(namespace) == 0 or namespace[0] == "": namespace = False
    profile_name = request.GET.getlist("profile_name")
    if len(profile_name) == 0 or profile_name[0] == "": profile_name = False
    start_time = request.GET.get("start_time")
    end_time = request.GET.get("end_time")
    type = request.GET.get("type")
    group_type = request.GET.getlist("group_type")
    if len(group_type) == 0 or group_type[0] == "": group_type = False
    group_name = request.GET.getlist("group_name")
    if len(group_name) == 0 or group_name[0] == "": group_name = False    
    service_flavour = request.GET.getlist("service_flavour")
    if len(service_flavour) == 0 or service_flavour[0] == "": service_flavour = False
    service_hostname = request.GET.getlist("service_hostname")
    if len(service_hostname) == 0 or service_hostname[0] == "": service_hostname = False
    output = lower(request.GET.get("output") or "xml")
    # Body
    if not vo_name or not profile_name or not start_time or not end_time\
            or not settings.ACE or not type in ['HOURLY', 'DAILY', 'WEEKLY', 'MONTHLY']:
        return HttpResponseBadRequest(error(output,3))
    results = {}
    if type == 'HOURLY':
        service_table = ACE_H_AVAILABILITY;
    elif type == 'DAILY':
        service_table = ACE_D_AVAILABILITY;
    elif type == 'WEEKLY':
        service_table = ACE_W_AVAILABILITY;
    elif type == 'MONTHLY':
        service_table = ACE_M_AVAILABILITY;
    else:
        return HttpResponseBadRequest(error(output,3))
    
    format = '%Y-%m-%dT%H:%M:%SZ'
    try:
       start_time_e = datetime(*time.strptime(start_time, format)[0:5]).strftime("%Y-%m-%d %H:%M:%S")
       end_time_e = datetime(*time.strptime(end_time, format)[0:5]).strftime("%Y-%m-%d %H:%M:%S")
    except ValueError, (ve):
       return HttpResponseBadRequest(error(output,"400: Bad Request. " + str(ve)))
    
    sfound=0
    service_ids=VoServiceGroup.objects
    if group_name:
       service_ids = service_ids.filter(groups__groupname__in = group_name)
       sfound=1  
    if group_type:
       service_ids = service_ids.filter(groups__group_type__typename__in = group_type)
       sfound=1
     
    profiles = Profile.objects.filter(name__in=profile_name,isdeleted='N')

    if namespace:
        profiles = profiles.filter(namespace__name__in=namespace)

    profiles = profiles.values_list('name','namespace__name','atp_vo__voname')\
                        .order_by('atp_vo__voname','name')
    count = 0
    for profile,namespace,vo in profiles:
        if not profile in results:
           results[profile]={}
        results[profile][namespace] = {
            'vo': vo,
            'services': {}
        }
        services = service_table.objects.filter(
            PROFILE__name = profile,
            PROFILE__namespace__name = namespace,
            TIMESTAMP__gte = start_time_e,
            TIMESTAMP__lte = end_time_e,
        )
        if service_flavour:
            services = services.filter(service__flavour__flavourname__in = service_flavour)
        if service_hostname:
            services = services.filter(service__hostname__in = service_hostname)
        if sfound:
            service_ids = service_ids.values_list('service',flat=True).distinct('service')
            services = services.filter(service__in = service_ids)
        count = count + services.count()
        if count>settings.DB_LIMIT:
            return HttpResponseBadRequest(error(output,1))
        
        services = services.values(
                'service__id',
                'service__hostname',
                'service__flavour__flavourname',
                'service__flavour__flavourname',
                'TIMESTAMP',
                'AVAILABILITY',
                'RELIABILITY',
                'MAINTENANCE',
            )
        results[profile][namespace]['services'] = {}
        for service in services:
                 if not service['service__id'] in results[profile][namespace]['services']:
                     results[profile][namespace]['services'][service['service__id']] = {
                            'hostname': service['service__hostname'],
                            'service_type': service['service__flavour__flavourname'],
                            'service_flavour': service['service__flavour__flavourname'],
                            'availabilities': []
                     }
                 results[profile][namespace]['services'][service['service__id']]['availabilities'].append({
                            'timestamp': service['TIMESTAMP'].isoformat() + 'Z',
                            'availability': service['AVAILABILITY'],
                            'reliability': service['RELIABILITY'],
                            'maintenance': service['MAINTENANCE']
                     })

    # Output
    if output == "json":
                return render_to_response(
            		'ws_service_availability_in_profile_json',
            		{'results': results},
            		mimetype = 'application/json',
        	)
    else:
        	return render_to_response(
            		'ws_service_availability_in_profile_xml',
            		{'results': results},
            		mimetype = 'application/xml',
        	)

# Group Availability in Profile
# /sam-pi/group_availability_in_profile?[vo_name]&[namespace]&[profile_name]&[start_time]&[end_time]&[type]&[group_type]&[group_name]&[output]
# vo_name: Mandatory and may appear more than once. Eg: cms
# namespace: Optional and may appear more than once. Eg: ch.cern.sam
# profile_name: Mandatory and may appear more than once. Eg: CMS_CRITICAL
# start_time: Mandatory. UTC time in W3C format: 'YYYY-MM-DDThh:mm:ssZ'. Eg: 2010-10-01T13:22:15Z
# end_time: Mandatory. UTC time in W3C format: 'YYYY-MM-DDThh:mm:ssZ'. Eg: 2010-10-02T00:00:00Z
# type: Mandatory. Possible values: 'HOURLY', 'DAILY', 'WEEKLY'.
# group_type: Optional and may appear more than once. Eg: Site
# group_name: Optional and may appear more than once. Eg:
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def group_availability_in_profile(request):
    # Params
    vo_name = request.GET.getlist("vo_name")
    if len(vo_name) == 0 or vo_name[0] == "": vo_name = False
    namespace = request.GET.getlist("namespace")
    if len(namespace) == 0 or namespace[0] == "": namespace = False
    profile_name = request.GET.getlist("profile_name")
    if len(profile_name) == 0 or profile_name[0] == "": profile_name = False
    start_time = request.GET.get("start_time")
    end_time = request.GET.get("end_time")
    type = request.GET.get("type")
    group_type = request.GET.getlist("group_type")
    if len(group_type) == 0 or group_type[0] == "": group_type = False
    group_name = request.GET.getlist("group_name")
    if len(group_name) == 0 or group_name[0] == "": group_name = False
    output = lower(request.GET.get("output") or "xml")
    # Body
    if not vo_name or not profile_name or not start_time or not end_time\
            or not settings.ACE \
            or not type in ['HOURLY', 'DAILY', 'WEEKLY', 'MONTHLY']:
        return HttpResponseBadRequest(error(output,3))
    results = {}
    if type == 'HOURLY':
        group_table = ACE_H_SITE_AVAILABILITY;
    elif type == 'DAILY':
        group_table = ACE_D_SITE_AVAILABILITY;
    elif type == 'WEEKLY':
        group_table = ACE_W_SITE_AVAILABILITY;
    elif type == 'MONTHLY':
        group_table = ACE_M_SITE_AVAILABILITY;
    else:
        return HttpResponseBadRequest(error(output,3))

    format = '%Y-%m-%dT%H:%M:%SZ'
    try:
       start_time_e = datetime(*time.strptime(start_time, format)[0:5]).strftime("%Y-%m-%d %H:%M:%S")
       end_time_e = datetime(*time.strptime(end_time, format)[0:5]).strftime("%Y-%m-%d %H:%M:%S")
    except ValueError, (ve):
       return HttpResponseBadRequest(error(output,"400: Bad Request. " + str(ve)))
    
    profiles = Profile.objects.filter(name__in=profile_name,isdeleted='N')

    if namespace:
        profiles = profiles.filter(namespace__name__in=namespace)

    profiles = profiles.values_list('name','namespace__name','atp_vo__voname')\
                        .order_by('atp_vo__voname','name')
    count = 0
    for profile,namespace,vo in profiles:
        if not profile in results:
           results[profile]={}

        results[profile][namespace] = {
            'vo': vo,
            'groups': {}
        }
        groups = group_table.objects.filter(
          PROFILE__name = profile,
          PROFILE__namespace__name = namespace,
          TIMESTAMP__gte = start_time_e,
          TIMESTAMP__lte = end_time_e,
        )
        if group_type:
           groups = groups.filter(GROUPING__group_type__typename__in = group_type)
        if group_name:
           groups = groups.filter(GROUPING__groupname__in = group_name)

        count = count + groups.count()
        if count>settings.DB_LIMIT:
            return HttpResponseBadRequest(error(output,1))

        groups = groups.order_by('TIMESTAMP')
 
        groups = groups.values(
          	'GROUPING__id',
            	'GROUPING__groupname',
            	'GROUPING__group_type__typename',
            	'TIMESTAMP',
            	'AVAILABILITY',
            	'RELIABILITY',
            	'MAINTENANCE',
            )
        results[profile][namespace]['groups'] = {}
        for group in groups:
            	if not group['GROUPING__id'] in results[profile][namespace]['groups']:
                	results[profile][namespace]['groups'][group['GROUPING__id']] = {
                    		'name': group['GROUPING__groupname'],
                    		'type': group['GROUPING__group_type__typename'],
                    		'availabilities': []
                	}	
            	results[profile][namespace]['groups'][group['GROUPING__id']]['availabilities'].append({
                		'timestamp': group['TIMESTAMP'].isoformat() + 'Z',
                		'availability': group['AVAILABILITY'],
                		'reliability': group['RELIABILITY'],
                		'maintenance': group['MAINTENANCE']
            		})

    
    # Output
    if output == "json":
        return render_to_response(
            'ws_group_availability_in_profile_json',
            {'results': results},
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_group_availability_in_profile_xml',
            {'results': results},
            mimetype = 'application/xml',
        )


# SLS Latest Site Availability
# /sam-pi/sls_latest_site_availability?[profile_name]&[site_name]&[output]
# profile_name: Mandatory. Eg: ROC_CRITICAL
# site_name: Mandatory. Eg: HG-06-EKT
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD) 
def sls_latest_site_availability(request):
    profile_name = request.GET.get("profile_name")
    site_name = request.GET.get("site_name")
    output = lower(request.GET.get("output") or "xml")
    if not site_name or not profile_name:
        return HttpResponseBadRequest(error(output, 3))

    # care only for known site statuses
    known_statuses = ['OK', 'WARNING', 'CRITICAL']

    try:
        # get profile from first available namespace
        profile = Profile.objects.filter(name=profile_name, isdeleted='N')[0]
    except IndexError, err:
        return HttpResponseNotFound(error(output, "404: Page Not Found." + str(err)))

    data = AceSiteStatus.objects.filter(
        profile=profile,
        grouping__groupname=site_name,
        grouping__group_type__typename=get_grouptype(profile.atp_vo.voname),
        metricstatus__description__in=known_statuses
    ).order_by('-timestamp').values_list(
        'grouping__groupname',
        'metricstatus__description',
        'end_timestamp',
        'downtime_flag'
    )

    try:
        (sitename, status, timestamp, downtime) = data[0]
        status_avail = {'OK': 100, 'WARNING': 100, 'CRITICAL': 0}
        avail = status_avail.get(status, 0)
    except IndexError:
        avail = -1
        sitename = site_name
        timestamp = datetime.utcnow()

    def utc_offset():
        ts = time.time()
        now, utcnow = datetime.fromtimestamp(ts), datetime.utcfromtimestamp(ts)
        return (now - utcnow).seconds // 3600
    # adjust utc time from db to local time
    timestamp += timedelta(hours=utc_offset())

    results = {
        'sitename': sitename,
        'availability': avail,
        'timestamp': "%s+0%d:00" % (timestamp.strftime('%Y-%m-%dT%H:%M:%S'), utc_offset())
    }

    if output == "json":
        return render_to_response(
            'ws_sls_latest_site_availability_json',
            results,
            mimetype = 'application/json',
        )
    return render_to_response(
        'ws_sls_latest_site_availability_xml',
        results,
        mimetype = 'application/xml',
    )


# C5 Monitored By SAM
# /sam-pi/c5_monitored_by_SAM?
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD)
def c5_monitored_by_SAM(request):
    lastweek = datetime.utcnow() - timedelta(days=7)
    lastweek_ts = time.mktime(lastweek.timetuple())
    monitored = MetricDataLatest.objects.filter(insert_time__gte=lastweek_ts)
    infrastructures = monitored.values('service__site__infrast',
                                       'service__site__infrast__infrastname').distinct()
    sites = {}
    for inf in infrastructures:
        infid = inf.get('service__site__infrast')
        infname = inf.get('service__site__infrast__infrastname')
        infsites = monitored.filter(service__site__infrast=infid)
        total = infsites.values('service__site').distinct().count()

        # get production sites monitored for this infrastructure
        production = infsites.filter(service__site__infrasttype='Production')
        production = production.values('service__site').distinct().count()

        # get certified sites monitored for this infrastructure
        certified = infsites.filter(service__site__certifstatus='Y')
        certified = certified.values('service__site').distinct().count()

        sites[infname] = {
            'total': total,
            'production': production,
            'certified': certified,
        }

    allservices = monitored.filter(service__isdeleted='N')
    services = {}
    services['total'] = allservices.values('service').distinct().count()
    services['production'] = allservices.filter(service__isinproduction='Y').values('service').distinct().count()

    nmetrics = monitored.values('metric').distinct().count()
    nvos = monitored.exclude(vo__isnull=True).values('vo').distinct().count()

    return render_to_response(
        'ws_c5_monitored_by_SAM_xml',
        {'sites': sites, 'services': services, 'nmetrics': nmetrics, 'nvos': nvos},
        mimetype='application/xml'
    )


# C5 Status of Sites Monitored by SAM
# /sam-pi/c5_status_of_sites_monitored_by_SAM?
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD)
def c5_status_of_sites_monitored_by_SAM(request):
    # computed site status for ops sites
    sitestatus = AceSiteStatusLatest.objects.filter(
        profile__name='WLCG_CREAM_LCGCE_CRITICAL',
        grouping__group_type__typename='Site'
    )

    # results for the last seven days
    lastweek = datetime.utcnow() - timedelta(days=7)
    lastweek_ts = time.mktime(lastweek.timetuple())
    monitored = MetricDataLatest.objects.filter(insert_time__gte=lastweek_ts)

    # all ops sites
    ops_sitegroups = Groups.objects.filter(group_type__typename='Site')
    statuses = {}
    simple_status = {'OK': 'up', 'WARNING': 'up', 'CRITICAL': 'down'}
    for infrast in Infrastructure.objects.all():
        # sites for certain inf that have been monitored in the last days
        infsites = Site.objects.filter(
            infrast=infrast, isdeleted='N', certifstatus='Y', infrasttype='Production',
            service__metricdatalatest__in=monitored
        ).distinct()
        statuses[infrast] = {'up': 0, 'down': 0, 'unknown': 0}
        for site in infsites:
            group = ops_sitegroups.filter(groupname=site.sitename)
            status = sitestatus.filter(grouping__in=group)
            if len(status):
                try:
                    s = simple_status.get(status[0].metricstatus.description, 'unknown')
                except ObjectDoesNotExist:
                    # some site statuses can not be mapped to metricstatus
                    continue
                statuses[infrast][s] += 1

    return render_to_response(
        'ws_c5_status_of_sites_monitored_by_SAM_xml',
        {'statuses': statuses},
        mimetype='application/xml'
    )


# C5 Software Versions
# /sam-pi/c5_software_versions?
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD)
def c5_software_versions(request):
    releases = {
        'glite': {
            'infrastructure': 'EGI',
            'metric_name': 'emi.wn.WN-SoftVer',
            'regex': r'glite-version\\n\+ mwver=.*Deducted middleware version: ([0-9.]+)'
        },
        'emi': {
            'infrastructure': 'EGI',
            'metric_name': 'emi.wn.WN-SoftVer',
            'regex': r'emi-version\\n\+ mwver=.*Deducted middleware version: ([0-9.]+)'
        },
        'osg': {
            'infrastructure': 'OSG',
            'metric_name': 'org.osg.general.osg-version',
            'regex': r'OSG ([0-9.]+)'
        },
    }

    lastweek = datetime.utcnow() - timedelta(days=7)
    lastweek_ts = time.mktime(lastweek.timetuple())

    versions = {}
    for name, query in releases.items():
        infrast = query.get('infrastructure')
        if infrast not in versions:
            versions[infrast] = {}
        versions[infrast][name] = {}

        sites = Site.objects.filter(
            isdeleted='N', certifstatus='Y', infrasttype='Production',
            infrast__infrastname=infrast
        )
        # fetch software version metric result for each site
        for site in sites:
            try:
                metric = MetricDataLatest.objects.filter(
                    metric__name=query.get('metric_name'),
                    metricdetail__detail__iregex=query.get('regex'),
                    service__site__sitename=site.sitename,
                    insert_time__gte=lastweek_ts
                ).order_by('-metricdetail')[0]
                sw_regex = re.compile(query.get('regex'), re.I)
                ver = sw_regex.findall(metric.metricdetail.detail)[0]
            except IndexError:
                # site has no records for this metric or version number not found
                continue

            if ver in versions[infrast][name]:
                versions[infrast][name][ver] += 1
            else:
                versions[infrast][name][ver] = 1
    
    return render_to_response(
        'ws_c5_software_versions_xml',
        {'versions': versions},
        mimetype='application/xml',
    )


# Metricstore Current Load per Minute
# /sam-pi/metricstore_current_load_per_minute?[hour]&[minute]&[output]
# hour: Optional. Eg: '10', '22'.
# minute: Optional. Eg: '4', '59'.
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD)
def metricstore_current_load_per_minute(request):
    # Params
    hour = request.GET.get("hour")
    minute = request.GET.get("minute")
    output = lower(request.GET.get("output") or "xml")
    # Body
    values={}
    metricresult = MetricstoreCurrLoadM.objects
    if hour:
       metricresult = metricresult.filter(hour=hour)
    if minute:
       metricresult = metricresult.filter(minute=minute)  
    values['results'] = {'field1':'Hour','field2':'Minute','data':[]}
    values['results']['data'] =	metricresult.values_list(
					'hour',
					'minute',
					'number_of_records')\
				.order_by('hour','minute')
    # Output
    if output == "json":
        return render_to_response(
            'ws_metricstore_current_load_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_metricstore_current_load_xml',
            values,
            mimetype = 'application/xml',
        )

# Metricstore Current Load per Hour
# /sam-pi/meitricstore_current_load_per_hour?[rec_date]&[hour]&[output]
# rec_date: Optional. Eg: '16-AUG-2011'.
# hour: Optional. Eg: '10', '22'.
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD)
def metricstore_current_load_per_hour(request):
    # Params
    rec_date = request.GET.get("rec_date")
    hour = request.GET.get("hour")
    output = lower(request.GET.get("output") or "xml")
    # Body
    values={}
    metricresult = MetricstoreCurrLoadH.objects
    if rec_date:
       metricresult = metricresult.filter(rec_date=rec_date)
    if hour:
       metricresult = metricresult.filter(hour=hour)
    values['results'] = {'field1':'Record_Date','field2':'Hour','data':[]}
    values['results']['data'] =  metricresult.values_list(
					'rec_date',
					'hour',
					'number_of_records')\
                                .order_by('rec_date','hour')
    # Output
    if output == "json":
        return render_to_response(
            'ws_metricstore_current_load_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_metricstore_current_load_xml',
            values,
            mimetype = 'application/xml',
        )

# Metricstore Current Load per Hour Detailed
# /sam-pi/metricstore_current_load_per_hour_detailed?[namespace]&[profile_name]&[ngi_name]&[flavour_name]&[rec_date]&[hour]&[output]
# namespace: Optional and may appear more than once. Eg: ch.cern.sam
# profile_name: Mandatory and may appear more than once. Eg: ROC_CRITICAL
# ngi_name: Optional and may appear more than once. Eg: NGI_IT
# flavour_name: Optional and may appear more than once. Eg: CREAM-CE
# rec_date: Optional. Eg: '16-AUG-2011'.
# hour: Optional. Eg: '10', '22'.
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD)
def metricstore_current_load_per_hour_detailed(request):
    # Params
    namespace = request.GET.getlist("namespace")
    if len(namespace) == 0 or namespace[0] == "": namespace = False
    profile_name = request.GET.getlist("profile_name")
    if len(profile_name) == 0 or profile_name[0] == "": profile_name = False
    ngi_name = request.GET.getlist("ngi_name")
    if len(ngi_name) == 0 or ngi_name[0] == "": ngi_name = False
    flavour_name = request.GET.getlist("flavour_name")
    if len(flavour_name) == 0 or flavour_name[0] == "": flavour_name = False
    rec_date = request.GET.get("rec_date")
    hour = request.GET.get("hour")
    output = lower(request.GET.get("output") or "xml")
    # Body
    if not profile_name: 
        return HttpResponseBadRequest(error(output,3))

    values={}
    values['results'] = {}
    profiles = Profile.objects.filter(name__in=profile_name,isdeleted='N')

    if namespace:
        profiles = profiles.filter(namespace__name__in=namespace)

    profiles = profiles.values_list('name','namespace__name','atp_vo__voname')\
                        .order_by('atp_vo__voname','name')
    count = 0
    for profile,namespace,vo in profiles:
         if not profile in values['results']:
             values['results'][profile]={}
         values['results'][profile][namespace]={}
         metricresult = MetricstoreCurrLoadHDET.objects.filter(
					profile__name = profile,
					profile__namespace__name = namespace)
         if ngi_name:
            metricresult = metricresult.filter(ngi_name__in=ngi_name)
         if flavour_name:
            metricresult = metricresult.filter(flavour__flavourname__in=flavour_name)
         if rec_date:
            metricresult = metricresult.filter(rec_date=rec_date)
         if hour:
            metricresult = metricresult.filter(hour=hour)
         metricresult = metricresult.values(
                                        'ngi_name',
                                        'flavour__flavourname',
                                        'rec_date',
                                        'hour',
                                        'number_of_records')\
                                .order_by('ngi_name','flavour__flavourname','rec_date','hour')
         for mresult in metricresult:
             if mresult['ngi_name'] not in values['results'][profile][namespace]:
                values['results'][profile][namespace][mresult['ngi_name']]={}
             if mresult['flavour__flavourname'] not in values['results'][profile][namespace][mresult['ngi_name']]:
                values['results'][profile][namespace][mresult['ngi_name']][mresult['flavour__flavourname']]=[]
             values['results'][profile][namespace][mresult['ngi_name']][mresult['flavour__flavourname']].append(
				(mresult['rec_date'],mresult['hour'],mresult['number_of_records']))
             
    # Output
    if output == "json":
        return render_to_response(
            'ws_metricstore_current_load_det_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_metricstore_current_load_det_xml',
            values,
            mimetype = 'application/xml',
        )


# Information about the release deployment
# /sam-pi/release_info
@cache_page(3600)
def release_info(request):
    return HttpResponse(
        simplejson.dumps(base.release_info()),
        mimetype='application/json'
    )


# Number Of Records in metricdata_spool
# /sam-pi/num_records_metricdata_spool?[output]
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD)
def num_records_metricdata_spool(request):
    # Params
    output = lower(request.GET.get("output") or "xml")
    # Body
    values={}
    values['results'] =  MetricdataSpool.objects.count()
    # Output
    if output == "json":
        return render_to_response(
            'ws_num_records_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_num_records_xml',
            values,
            mimetype = 'application/xml',
        )

# Number Of Records in metricdataforrecalculation
# /sam-pi/num_records_metricdataforrecalculation?[output]
# output: Optional. Possible values are: XML, JSON. Default: XML.
@cache_page(60)
@throttle(settings.NUMBER_OF_ACCESSES, settings.ACCESS_PERIOD)
def num_records_metricdataforrecalculation(request):
    # Params
    output = lower(request.GET.get("output") or "xml")
    # Body
    values={}
    try:
        values['results'] =  MetricdataForRecalculation.objects.count()
    except DatabaseError:
        values['results'] = 0
    # Output
    if output == "json":
        return render_to_response(
            'ws_num_records_json',
            values,
            mimetype = 'application/json',
        )
    else:
        return render_to_response(
            'ws_num_records_xml',
            values,
            mimetype = 'application/xml',
        )

# Information about running ACE computations
# /sam-pi/ace_status?[output]
# output: Optional. Possible values are: XML, JSON. Default: XML.
def ace_status(request):
    output = lower(request.GET.get("output", ''))
    if output not in ('json', 'xml'):
        output = 'xml'

    # last status calculation dates
    service = AceServiceStatus.objects.aggregate(Max('end_timestamp'))
    site = AceSiteStatus.objects.aggregate(Max('end_timestamp'))
    # availability tasks not yet completed
    tasks = AceAvailabilityTask.objects.exclude(considered='Y')
    data = {'last_service_status_date': service.get('end_timestamp__max'),
            'last_site_status_date': site.get('end_timestamp__max'),
            'availability_tasks': tasks.order_by('-end_timestamp')}

    template = 'ws_ace_status_%s' % output
    mimetype = 'application/%s' % output
    return render_to_response(template, data, mimetype=mimetype)


# Number of FTS records arrived in the last hour
# /sam-pi/fts_records_arrived?[output]
# ouutput: Optional. Possible values are: XML, JSON. Default: XML
def fts_records_arrived(request):
    output = lower(request.GET.get("output", ''))
    if output not in ('json', 'xml'):
        output = 'xml'

    # one hour range (non including) in milliseconds
    end = 1000 * time.mktime(datetime.utcnow().timetuple()) + 1
    start = 1000 * (end - 3600) - 1
    records = FtsDataRaw.objects.filter(timestamp_tr_comp__range=(start, end))
    data = {'records': records.count()}

    template = 'ws_fts_records_%s' % output
    mimetype = 'application/%s' % output
    return render_to_response(template, data, mimetype=mimetype)

# Number of FTS records summarized in the last hour
# /sam-pi/fts_records_summarized?[output]
# ouutput: Optional. Possible values are: XML, JSON. Default: XML
def fts_records_summarized(request):
    output = lower(request.GET.get("output", ''))
    if output not in ('json', 'xml'):
        output = 'xml'
    
    end = time.mktime(datetime.utcnow().timetuple()) + 1
    start = (end - 3600) - 1
    records = FtsHourlySummary.objects.filter(timestamp__range=(start, end))
    data = {'records': records.count()}

    template = 'ws_fts_records_%s' % output
    mimetype = 'application/%s' % output
    return render_to_response(template, data, mimetype=mimetype)

    
